/*!
        @file    $Id:: field.cpp #$

        @brief

        @author  <Hideo Matsufuru> hideo.matsufuru@kek.jp(matsufuru)
                 $LastChangedBy: aoym $

        @date    $LastChangedDate:: 2014-09-16 12:23:39 #$

        @version $LastChangedRevision: 1144 $
*/

#include "field.h"
#include <cstring>

#include "threadManager_OpenMP.h"

#include "bridgeIO.h"
using Bridge::vout;

using std::string;

//====================================================================
namespace {
  inline
  void set_threadtask(int& ith, int& nth, int& is, int& ns,
                      const int size)
  {
    nth = ThreadManager_OpenMP::get_num_threads();
    ith = ThreadManager_OpenMP::get_thread_id();

    is = size * ith / nth;
    ns = size * (ith + 1) / nth;
  }
}

//====================================================================
void Field::check()
{
  //    vout.general("Field was constructed.\n");
}


//====================================================================
double dot(const Field& y, const Field& x)
{
  double *yp = const_cast<Field *>(&y)->ptr(0);
  double *xp = const_cast<Field *>(&x)->ptr(0);

  //  int size = x.ntot();
  int ith, nth, is, ns;

  set_threadtask(ith, nth, is, ns, x.ntot());

  double a = 0.0;

  //  for (int k = 0; k < size; ++k) {
  for (int k = is; k < ns; ++k) {
    a += yp[k] * xp[k];
  }

  //  double b = Communicator::reduce_sum(a);
  //  return b;
  ThreadManager_OpenMP::reduce_sum_global(a, ith, nth);
  return a;
}


//====================================================================
double dot(const Field& y, const int exy, const Field& x, const int exx)
{
  assert(x.nin() == y.nin());
  assert(x.nvol() == y.nvol());

  double *yp = const_cast<Field *>(&y)->ptr(0, 0, exy);
  double *xp = const_cast<Field *>(&x)->ptr(0, 0, exx);

  //  int size = x.nin() * x.nvol();
  int ith, nth, is, ns;
  set_threadtask(ith, nth, is, ns, x.ntot());

  double a = 0.0;

  //  for (int k = 0; k < size; ++k) {
  for (int k = is; k < ns; ++k) {
    a += yp[k] * xp[k];
  }

  //  double b = Communicator::reduce_sum(a);
  //  return b;
  ThreadManager_OpenMP::reduce_sum_global(a, ith, nth);
  return a;
}


//====================================================================
dcomplex dotc(const Field& y, const Field& x)
{
  double *yp = const_cast<Field *>(&y)->ptr(0);
  double *xp = const_cast<Field *>(&x)->ptr(0);

  assert(x.ntot() == y.ntot());

  //  int ntot = x.ntot();
  int ith, nth, is, ns;
  set_threadtask(ith, nth, is, ns, x.ntot());

  if ((y.field_element_type() == Field::COMPLEX) &&
      (x.field_element_type() == Field::COMPLEX)) {
    double prdr = 0.0;
    double prdi = 0.0;

    //    for (int k = 0; k < ntot; k += 2) {
    for (int k = is; k < ns; k += 2) {
      prdr += yp[k] * xp[k] + yp[k + 1] * xp[k + 1];
      prdi += yp[k] * xp[k + 1] - yp[k + 1] * xp[k];
    }

    //    double prd[2];
    //    prd[0] = prdr;
    //    prd[1] = prdi;

    //    double prd_g[2];
    //    Communicator::reduce_sum(2, prd_g, prd);

    //    return cmplx(prd_g[0], prd_g[1]);

    ThreadManager_OpenMP::reduce_sum_global(prdr, ith, nth);
    ThreadManager_OpenMP::reduce_sum_global(prdi, ith, nth);

    return cmplx(prdr, prdi);
  } else if ((y.field_element_type() == Field::REAL) &&
             (y.field_element_type() == Field::REAL)) {
    return cmplx(dot(y, x), 0.0);
  } else {
    vout.crucial("%s: unsupported arg types.\n", __func__);
    abort();
    return cmplx(0.0, 0.0);  // never reached.
  }
}


//====================================================================
dcomplex dotc(const Field& y, const int exy, const Field& x, const int exx)
{
  double *yp = const_cast<Field *>(&y)->ptr(0, 0, exy);
  double *xp = const_cast<Field *>(&x)->ptr(0, 0, exx);

  assert(x.nin() == y.nin());
  assert(x.nvol() == y.nvol());

  //  int size = x.nin() * x.nvol();
  int ith, nth, is, ns;
  set_threadtask(ith, nth, is, ns, x.nin() * x.nvol());

  if ((y.field_element_type() == Field::COMPLEX) &&
      (x.field_element_type() == Field::COMPLEX)) {
    double prdr = 0.0;
    double prdi = 0.0;

    //    for (int k = 0; k < size; k += 2) {
    for (int k = is; k < ns; k += 2) {
      prdr += yp[k] * xp[k] + yp[k + 1] * xp[k + 1];
      prdi += yp[k] * xp[k + 1] - yp[k + 1] * xp[k];
    }

    //    double prd[2];
    //    prd[0] = prdr;
    //    prd[1] = prdi;

    //    double prd_g[2];
    //    Communicator::reduce_sum(2, prd_g, prd);

    //    return cmplx(prd_g[0], prd_g[1]);

    ThreadManager_OpenMP::reduce_sum_global(prdr, ith, nth);
    ThreadManager_OpenMP::reduce_sum_global(prdi, ith, nth);

    return cmplx(prdr, prdi);
  } else if ((y.field_element_type() == Field::REAL) &&
             (y.field_element_type() == Field::REAL)) {
    return cmplx(dot(y, exy, x, exx), 0.0);
  } else {
    vout.crucial("%s: unsupported arg types.\n", __func__);
    abort();
    return cmplx(0.0, 0.0);  // never reached.
  }
}


//====================================================================
void axpy(Field& y, const double a, const Field& x)
{
  double *yp = const_cast<Field *>(&y)->ptr(0);
  double *xp = const_cast<Field *>(&x)->ptr(0);

  //  int size = x.ntot();
  int ith, nth, is, ns;

  set_threadtask(ith, nth, is, ns, x.ntot());

  //  for (int k = 0; k < size; ++k) {
  for (int k = is; k < ns; ++k) {
    yp[k] += a * xp[k];
  }
}


//====================================================================
void axpy(Field& y, const int exy, const double a, const Field& x, const int exx)
{
  assert(x.nin() == y.nin());
  assert(x.nvol() == y.nvol());

  double *yp = const_cast<Field *>(&y)->ptr(0, 0, exy);
  double *xp = const_cast<Field *>(&x)->ptr(0, 0, exx);

  //  int size = x.nin() * x.nvol();
  int ith, nth, is, ns;
  set_threadtask(ith, nth, is, ns, x.nin() * x.nvol());

  //  for (int k = 0; k < size; ++k) {
  for (int k = is; k < ns; ++k) {
    yp[k] += a * xp[k];
  }
}


//====================================================================
void axpy(Field& y, const dcomplex a, const Field& x)
{
  if (imag(a) == 0.0) {
    return axpy(y, real(a), x);
  } else if ((y.field_element_type() == Field::REAL) &&
             (x.field_element_type() == Field::REAL)) {
    vout.crucial("%s: real vector and complex parameter.\n", __func__);
    abort();

    return axpy(y, real(a), x);  // ignore imaginary part of a
  } else if ((y.field_element_type() == Field::COMPLEX) &&
             (x.field_element_type() == Field::COMPLEX)) {
    double *yp = const_cast<Field *>(&y)->ptr(0);
    double *xp = const_cast<Field *>(&x)->ptr(0);

    assert(x.ntot() == y.ntot());

    //    int ntot = x.ntot();
    int ith, nth, is, ns;
    set_threadtask(ith, nth, is, ns, x.ntot());

    double ar = real(a);
    double ai = imag(a);

    //    for (int k = 0; k < ntot; k += 2) {
    for (int k = is; k < ns; k += 2) {
      yp[k]     += ar * xp[k] - ai * xp[k + 1];
      yp[k + 1] += ar * xp[k + 1] + ai * xp[k];
    }

    return;
  } else {
    vout.crucial("%s: unsupported types.\n", __func__);
    abort();
  }
}


//====================================================================
void axpy(Field& y, const int exy, const dcomplex a, const Field& x, const int exx)
{
  if (imag(a) == 0.0) {
    return axpy(y, exy, real(a), x, exx);
  } else if ((y.field_element_type() == Field::REAL) &&
             (x.field_element_type() == Field::REAL)) {
    vout.crucial("%s: real vector and complex parameter.\n", __func__);
    abort();

    return axpy(y, real(a), x);  // ignore imaginary part of a
  } else if ((y.field_element_type() == Field::COMPLEX) &&
             (x.field_element_type() == Field::COMPLEX)) {
    double *yp = const_cast<Field *>(&y)->ptr(0, 0, exy);
    double *xp = const_cast<Field *>(&x)->ptr(0, 0, exx);

    assert(x.nin() == y.nin());
    assert(x.nvol() == y.nvol());

    //    int size = x.nin() * x.nvol();
    int ith, nth, is, ns;
    set_threadtask(ith, nth, is, ns, x.nin() * x.nvol());

    double ar = real(a);
    double ai = imag(a);

    //    for (int k = 0; k < size; k += 2) {
    for (int k = is; k < ns; k += 2) {
      yp[k]     += ar * xp[k] - ai * xp[k + 1];
      yp[k + 1] += ar * xp[k + 1] + ai * xp[k];
    }

    return;
  } else {
    vout.crucial("%s: unsupported types.\n", __func__);
    abort();
  }
}


//====================================================================
void scal(Field& x, const double a)
{
  //  x.field *= a;

  double *xp = const_cast<Field *>(&x)->ptr(0, 0, 0);
  int    ith, nth, is, ns;

  set_threadtask(ith, nth, is, ns, x.ntot());

  for (int k = is; k < ns; ++k) {
    xp[k] *= a;
  }
}


//====================================================================
void scal(Field& x, const int exx, const double a)
{
  double *xp = const_cast<Field *>(&x)->ptr(0, 0, exx);
  //  int size = x.nin() * x.nvol();
  int ith, nth, is, ns;

  set_threadtask(ith, nth, is, ns, x.nin() * x.nvol());

  //  for (int k = 0; k < size; ++k) {
  for (int k = is; k < ns; ++k) {
    //    x.field[k] *= a;
    xp[k] *= a;
  }
}


//====================================================================
void scal(Field& x, const dcomplex a)
{
  if (x.field_element_type() == Field::REAL) {
    vout.crucial("%s: real vector and complex parameter.\n", __func__);
    abort();

    //    x.field *= real(a);  // ignore imaginary part of a
    scal(x, a);
  } else if (x.field_element_type() == Field::COMPLEX) {
    double *xp = const_cast<Field *>(&x)->ptr(0);
    //int ntot = x.ntot();
    int ith, nth, is, ns;
    set_threadtask(ith, nth, is, ns, x.ntot());

    double ar = real(a);
    double ai = imag(a);

    //    for (int k = 0; k < ntot; k += 2) {
    for (int k = is; k < ns; k += 2) {
      double xr = xp[k];
      double xi = xp[k + 1];

      xp[k]     = ar * xr - ai * xi;
      xp[k + 1] = ar * xi + ai * xr;
    }
  } else {
    vout.crucial("%s: unsupported field type.\n", __func__);
    abort();
  }
}


//====================================================================
void scal(Field& x, const int exx, const dcomplex a)
{
  if (x.field_element_type() == Field::REAL) {
    vout.crucial("%s: real vector and complex parameter.\n", __func__);
    abort();

    //    x.field *= real(a);  // ignore imaginary part of a
    scal(x, exx, a);
  } else if (x.field_element_type() == Field::COMPLEX) {
    double *xp = const_cast<Field *>(&x)->ptr(0, 0, exx);
    //    int size = x.nin() * x.nvol();
    int ith, nth, is, ns;
    set_threadtask(ith, nth, is, ns, x.nin() * x.nvol());

    double ar = real(a);
    double ai = imag(a);

    //    for (int k = 0; k < size; k += 2) {
    for (int k = is; k < ns; k += 2) {
      double xr = xp[k];
      double xi = xp[k + 1];

      xp[k]     = ar * xr - ai * xi;
      xp[k + 1] = ar * xi + ai * xr;
    }
  } else {
    vout.crucial("%s: unsupported field type.\n", __func__);
    abort();
  }
}


//====================================================================
void copy(Field& y, const Field& x)
{
  //  y.field = x.field;

  assert(x.nin() == y.nin());
  assert(x.nvol() == y.nvol());
  assert(x.nex() == y.nex());

  double *yp = const_cast<Field *>(&y)->ptr(0, 0, 0);
  double *xp = const_cast<Field *>(&x)->ptr(0, 0, 0);

  int ith, nth, is, ns;
  set_threadtask(ith, nth, is, ns, x.ntot());

  for (int k = is; k < ns; ++k) {
    yp[k] = xp[k];
  }
}


//====================================================================
void copy(Field& y, const int exy, const Field& x, const int exx)
{
  assert(x.nin() == y.nin());
  assert(x.nvol() == y.nvol());

  double *yp = const_cast<Field *>(&y)->ptr(0, 0, exy);
  double *xp = const_cast<Field *>(&x)->ptr(0, 0, exx);

  //  int size = x.nin() * x.nvol();
  int ith, nth, is, ns;
  set_threadtask(ith, nth, is, ns, x.nin() * x.nvol());

  for (int k = is; k < ns; ++k) {
    yp[k] = xp[k];
  }

  ////  for (int k = 0; k < size; ++k) {
  ////    yp[k] = xp[k];
  ////  }
  //  memcpy(yp, xp, sizeof(double) * size);
}


//====================================================================
void Field::set(double a)
{
  int ith, nth, is, ns;

  set_threadtask(ith, nth, is, ns, ntot());

  double *yp = const_cast<Field *>(this)->ptr(0);

  for (int k = is; k < ns; ++k) {
    yp[k] = a;
  }
}


//====================================================================
double Field::norm2() const
{
  int ith, nth, is, ns;

  set_threadtask(ith, nth, is, ns, ntot());

  double *yp = const_cast<Field *>(this)->ptr(0);

  double a = 0.0;

  for (int k = is; k < ns; ++k) {
    a += yp[k] * yp[k];
  }

  ThreadManager_OpenMP::reduce_sum_global(a, ith, nth);
  return a;
}


//====================================================================
void aypx(const double a, Field& y, const Field& x)
{
  int ith, nth, is, ns;

  set_threadtask(ith, nth, is, ns, x.ntot());

  double *yp = const_cast<Field *>(&y)->ptr(0);
  double *xp = const_cast<Field *>(&x)->ptr(0);

  for (int k = is; k < ns; ++k) {
    yp[k] = a * yp[k] + xp[k];
  }
}


//====================================================================
void aypx(const dcomplex a, Field& y, const Field& x)
{
  if (imag(a) == 0.0) {
    return aypx(real(a), y, x);
  } else if ((y.field_element_type() == Field::REAL) &&
             (x.field_element_type() == Field::REAL)) {
    vout.crucial("%s: real vector and complex parameter.\n", __func__);
    abort();

    return aypx(real(a), y, x);  // ignore imaginary part of a
  } else if ((y.field_element_type() == Field::COMPLEX) &&
             (x.field_element_type() == Field::COMPLEX)) {
    double *yp = const_cast<Field *>(&y)->ptr(0);
    double *xp = const_cast<Field *>(&x)->ptr(0);

    assert(x.ntot() == y.ntot());

    int ith, nth, is, ns;
    set_threadtask(ith, nth, is, ns, x.ntot());

    double ar = real(a);
    double ai = imag(a);

    for (int k = is; k < ns; k += 2) {
      double ypr = yp[k];
      double ypi = yp[k + 1];
      yp[k]     = ar * ypr - ai * ypi + xp[k];
      yp[k + 1] = ar * ypi + ai * ypr + xp[k + 1];
    }

    return;
  } else {
    vout.crucial("%s: unsupported types.\n", __func__);
    abort();
  }
}


//====================================================================
void Field::stat(double& Fave, double& Fmax, double& Fdev) const
{
  assert(ThreadManager_OpenMP::get_num_threads() == 1);

  double sum  = 0.0;
  double sum2 = 0.0;

  Fmax = 0.0;
  for (int ex = 0, nnex = nex(); ex < nnex; ++ex) {
    for (int site = 0, nnvol = nvol(); site < nnvol; ++site) {
      double fst = 0.0;
      for (int in = 0, nnin = nin(); in < nnin; ++in) {
        double fv = field[myindex(in, site, ex)];
        fst += fv * fv;
      }
      sum2 += fst;
      fst   = sqrt(fst);
      sum  += fst;
      if (fst > Fmax) Fmax = fst;
    }
  }

  sum  = Communicator::reduce_sum(sum);
  sum2 = Communicator::reduce_sum(sum2);
  Fmax = Communicator::reduce_max(Fmax);

  double perdeg = 1.0 / ((double)CommonParameters::Lvol() * nex());
  Fave = sum * perdeg;

  double fval = sum2 * perdeg;
  fval -= Fave * Fave;
  Fdev  = sqrt(fval);
}


//====================================================================
void report_field_stat(const Bridge::VerboseLevel vl,
                       const std::string& msg, const Field& f)
{
  assert(ThreadManager_OpenMP::get_num_threads() == 1);

  double favg, fmax, fdev;
  f.stat(favg, fmax, fdev);

  vout.general(vl,
               "    %s: avg = %12.6f  max = %12.6f  dev = %12.6f\n",
               msg.c_str(),
               favg, fmax, fdev);
}


//====================================================================
//============================================================END=====
