/*!
        @file    $Id:: fopr_Clover_Isochemical.cpp #$

        @brief

        @author  <Hideo Matsufuru> hideo.matsufuru@kek.jp(matsufuru)
                 $LastChangedBy: aoym $

        @date    $LastChangedDate:: 2014-09-16 12:23:39 #$

        @version $LastChangedRevision: 1144 $
*/

#include "fopr_Clover_Isochemical.h"

#ifdef USE_PARAMETERS_FACTORY
#include "parameters_factory.h"
#endif

using std::valarray;
using std::string;

//- parameter entries
namespace {
  void append_entry(Parameters& param)
  {
    param.Register_string("gamma_matrix_type", "NULL");
    param.Register_double("hopping_parameter", 0.0);
    param.Register_double("clover_coefficient", 0.0);
    param.Register_double("isospin_chemical_potential", 0.0);
    param.Register_int_vector("boundary_condition", std::valarray<int>());

    param.Register_string("verbose_level", "NULL");
  }


#ifdef USE_PARAMETERS_FACTORY
  bool init_param = ParametersFactory::Register("Fopr.Clover_Isochemical", append_entry);
#endif
}
//- end

//- parameters class
Parameters_Fopr_Clover_Isochemical::Parameters_Fopr_Clover_Isochemical() { append_entry(*this); }
//- end

const std::string Fopr_Clover_Isochemical::class_name = "Fopr_Clover_Isochemical";

//====================================================================
void Fopr_Clover_Isochemical::set_parameters(const Parameters& params)
{
  const string str_vlevel = params.get_string("verbose_level");

  m_vl = vout.set_verbose_level(str_vlevel);

  //- fetch and check input parameters
  double        kappa, cSW, mu;
  valarray<int> bc;

  int err = 0;
  err += params.fetch_double("hopping_parameter", kappa);
  err += params.fetch_double("clover_coefficient", cSW);
  err += params.fetch_double("isospin_chemical_potential", mu);
  err += params.fetch_int_vector("boundary_condition", bc);

  if (err) {
    vout.crucial(m_vl, "%s: fetch error, input parameter not found.\n", class_name.c_str());
    abort();
  }


  set_parameters(kappa, cSW, mu, bc);
}


//====================================================================
void Fopr_Clover_Isochemical::set_parameters(double kappa, double cSW, double mu,
                                             valarray<int> bc)
{
  //- print input parameters
  vout.general(m_vl, "Parameters of %s:\n", class_name.c_str());
  vout.general(m_vl, "  kappa = %8.4f\n", kappa);
  vout.general(m_vl, "  cSW   = %8.4f\n", cSW);
  vout.general(m_vl, "  mu    = %8.4f\n", mu);
  for (int dir = 0; dir < m_Ndim; ++dir) {
    vout.general(m_vl, "  boundary[%d] = %2d\n", dir, bc[dir]);
  }

  //- range check
  // NB. kappa,cSW,mu == 0 is allowed.
  assert(bc.size() == m_Ndim);

  //- store values
  m_kappa = kappa;
  m_cSW   = cSW;
  m_mu    = mu;

  for (int dir = 0; dir < m_Ndim; ++dir) {
    m_boundary[dir] = bc[dir];
  }

  //- propagate parameters
  m_fopr_w->set_parameters(m_kappa, m_mu, m_boundary);
  m_fopr_csw->set_parameters(m_kappa, m_cSW, m_boundary);
}


//====================================================================
void Fopr_Clover_Isochemical::init(string repr)
{
  m_Nvol = CommonParameters::Nvol();
  m_Ndim = CommonParameters::Ndim();
  m_Nc   = CommonParameters::Nc();
  m_Nd   = CommonParameters::Nd();
  m_NinF = 2 * m_Nc * m_Nd;

  m_boundary.resize(m_Ndim);

  m_U = 0;

  m_repr = repr;

  m_fopr_w   = new Fopr_Wilson_Isochemical(repr);
  m_fopr_csw = new Fopr_CloverTerm(repr);

  m_w1.reset(m_NinF, m_Nvol, 1);
  m_w2.reset(m_NinF, m_Nvol, 1);
}


//====================================================================
void Fopr_Clover_Isochemical::tidyup()
{
  delete m_fopr_w;
  delete m_fopr_csw;
}


//====================================================================
void Fopr_Clover_Isochemical::set_config(Field *U)
{
  m_U = (Field_G *)U;

  m_fopr_w->set_config(U);
  m_fopr_csw->set_config(U);
}


//====================================================================
void Fopr_Clover_Isochemical::D(Field& w, const Field& f)
{
  assert(f.nex() == 1);
  //  Field w2(f.nin(), f.nvol(), 1);

  m_fopr_w->D(w, f);
  m_fopr_csw->mult_sigmaF(m_w1, f);
  axpy(w, -1.0, m_w1);  //  w -= m_w1;
}


//====================================================================
void Fopr_Clover_Isochemical::Ddag(Field& w, const Field& f)
{
  assert(f.nex() == 1);
  //  Field w2(f.nin(), f.nvol(), 1);

  m_fopr_w->Ddag(w, f);
  m_fopr_csw->mult_sigmaF(m_w1, f);
  axpy(w, -1.0, m_w1);  //  w -= m_w1;
}


//====================================================================
void Fopr_Clover_Isochemical::DdagD(Field& w, const Field& f)
{
  assert(f.nex() == 1);
  //  Field w2(f.nin(), f.nvol(), 1);

  D(m_w2, f);
  Ddag(w, m_w2);
}


//====================================================================
void Fopr_Clover_Isochemical::H(Field& w, const Field& f)
{
  assert(f.nex() == 1);
  //  Field w2(f.nin(), f.nvol(), 1);

  D(m_w2, f);
  mult_gm5(w, m_w2);
}


//====================================================================
void Fopr_Clover_Isochemical::Hdag(Field& w, const Field& f)
{
  assert(f.nex() == 1);
  //  Field w2(f.nin(), f.nvol(), 1);

  mult_gm5(m_w2, f);
  Ddag(w, m_w2);
}


//====================================================================
void Fopr_Clover_Isochemical::mult_isigma(Field_F& v, const Field_F& w,
                                          const int mu, const int nu)
{
  m_fopr_csw->mult_isigma(v, w, mu, nu);
}


//====================================================================
//============================================================END=====
