/*!
        @file    $Id:: force_F_Clover_Nf2.cpp #$

        @brief

        @author  <Hideo Matsufuru> hideo.matsufuru@kek.jp(matsufuru)
                 $LastChangedBy: namekawa $

        @date    $LastChangedDate:: 2014-09-16 18:27:22 #$

        @version $LastChangedRevision: 1145 $
*/

#include "force_F_Clover_Nf2.h"

#ifdef USE_PARAMETERS_FACTORY
#include "parameters_factory.h"
#endif

using std::valarray;

//- parameter entries
namespace {
  void append_entry(Parameters& param)
  {
    param.Register_double("hopping_parameter", 0.0);
    param.Register_double("clover_coefficient", 0.0);
    param.Register_int_vector("boundary_condition", std::valarray<int>());

    param.Register_string("verbose_level", "NULL");
  }


#ifdef USE_PARAMETERS_FACTORY
  bool init_param = ParametersFactory::Register("Force.F_Clover_Nf2", append_entry);
#endif
}
//- end

//- parameters class
Parameters_Force_F_Clover_Nf2::Parameters_Force_F_Clover_Nf2() { append_entry(*this); }
//- end

const std::string Force_F_Clover_Nf2::class_name = "Force_F_Clover_Nf2";

//====================================================================
void Force_F_Clover_Nf2::set_parameters(const Parameters& params)
{
  const string str_vlevel = params.get_string("verbose_level");

  m_vl = vout.set_verbose_level(str_vlevel);

  //- fetch and check input parameters
  double        kappa, cSW;
  valarray<int> bc;

  int err = 0;
  err += params.fetch_double("hopping_parameter", kappa);
  err += params.fetch_double("clover_coefficient", cSW);
  err += params.fetch_int_vector("boundary_condition", bc);

  if (err) {
    vout.crucial(m_vl, "%s: fetch error, input parameter not found.\n", class_name.c_str());
    abort();
  }


  set_parameters(kappa, cSW, bc);
}


//====================================================================
void Force_F_Clover_Nf2::set_parameters(const double kappa, const double cSW,
                                        const valarray<int> bc)
{
  int Ndim = CommonParameters::Ndim();

  //- print input parameters
  vout.general(m_vl, "Parameters of %s:\n", class_name.c_str());
  vout.general(m_vl, "  kappa = %8.4f\n", kappa);
  vout.general(m_vl, "  cSW   = %8.4f\n", cSW);
  for (int mu = 0; mu < Ndim; ++mu) {
    vout.general(m_vl, "  boundary[%d] = %2d\n", mu, bc[mu]);
  }

  //- range check
  // NB. kappa,cSW == 0 is allowed.
  assert(bc.size() == Ndim);

  //- store values
  m_kappa = kappa;
  m_cSW   = cSW;

  m_boundary.resize(Ndim);
  for (int mu = 0; mu < Ndim; ++mu) {
    m_boundary[mu] = bc[mu];
  }

  //- propagate parameters
  m_fopr_c->set_parameters(m_kappa, m_cSW, m_boundary);

  m_force_w->set_parameters(m_kappa, m_boundary);
  m_force_csw->set_parameters(m_kappa, m_cSW, m_boundary);
}


//====================================================================
void Force_F_Clover_Nf2::init(std::string repr)
{
  m_repr = repr;

  m_fopr_c    = new Fopr_Clover(repr);
  m_force_w   = new Force_F_Wilson_Nf2(repr);
  m_force_csw = new Force_F_CloverTerm(repr);

  m_boundary.resize(CommonParameters::Ndim());

  int Nvol = CommonParameters::Nvol();
  m_Ndim = CommonParameters::Ndim();

  m_Cud = new Field_G(Nvol, m_Ndim * m_Ndim);
}


//====================================================================
void Force_F_Clover_Nf2::tidyup()
{
  delete m_Cud;
  delete m_force_csw;
  delete m_force_w;
  delete m_fopr_c;
}


//====================================================================
void Force_F_Clover_Nf2::force_udiv(Field& force_, const Field& eta_)
{
  int Nc   = CommonParameters::Nc();
  int Nvol = CommonParameters::Nvol();
  int Ndim = CommonParameters::Ndim();

  Field_G force1(Nvol, Ndim);

  Field_F zeta(Nvol, 1);
  Field_F eta(eta_);

  m_fopr_c->H(zeta, eta);

  force_udiv1(force1, eta, zeta);
  force_ = force1;

  force_udiv1(force1, zeta, eta);
  force_ += force1;
}


//====================================================================
void Force_F_Clover_Nf2::force_udiv1(Field& force_, const Field& zeta_, const Field& eta_)
{
  int Nc   = CommonParameters::Nc();
  int Nvol = CommonParameters::Nvol();
  int Ndim = CommonParameters::Ndim();

  Field_G force(Nvol, Ndim);
  Field_F zeta(zeta_);
  Field_F eta(eta_);

  force_udiv1_impl(force, zeta, eta);

  force_ = force;
}


//====================================================================
void Force_F_Clover_Nf2::force_udiv1_impl(Field_G& force, const Field_F& zeta, const Field_F& eta)
{
  int Nc   = CommonParameters::Nc();
  int Nd   = CommonParameters::Nd();
  int Nvol = CommonParameters::Nvol();
  int Ndim = CommonParameters::Ndim();
  // int NinG = 2 * Nc * Nc;

  ShiftField_lex shift;

//  Field force(NinG, Nvol, Ndim);
  force = 0.0;

//  Field force2(NinG, Nvol, Ndim);
  Field_G force2(Nvol, Ndim);

  m_force_w->force_udiv1(force, zeta, eta);
  m_force_csw->force_udiv1(force2, zeta, eta);

  force += force2;
}


//====================================================================
void Force_F_Clover_Nf2::set_component()
{
  int Nc   = CommonParameters::Nc();
  int Nd   = CommonParameters::Nd();
  int Nvol = CommonParameters::Nvol();

  Staples staple;
  Field_G Cmu_ud(Nvol, 1);

  for (int mu = 0; mu < m_Ndim; ++mu) {
    for (int nu = 0; nu < m_Ndim; ++nu) {
      if (nu == mu) continue;

      Cmu_ud  = staple.upper(*m_U, mu, nu);
      Cmu_ud -= staple.lower(*m_U, mu, nu);
      m_Cud->setpart_ex(index_dir(mu, nu), Cmu_ud, 0);
    }
  }
}


//====================================================================
//============================================================END=====
