/*!
        @file    $Id:: force_F_Clover_Nf2_Isochemical.cpp #$

        @brief

        @author  <Hideo Matsufuru> hideo.matsufuru@kek.jp(matsufuru)
                 $LastChangedBy: namekawa $

        @date    $LastChangedDate:: 2014-09-16 18:27:22 #$

        @version $LastChangedRevision: 1145 $
*/

#include "force_F_Clover_Nf2_Isochemical.h"

#ifdef USE_PARAMETERS_FACTORY
#include "parameters_factory.h"
#endif

using std::valarray;

//- parameter entries
namespace {
  void append_entry(Parameters& param)
  {
    param.Register_double("hopping_parameter", 0.0);
    param.Register_double("clover_coefficient", 0.0);
    param.Register_int_vector("boundary_condition", std::valarray<int>());

    param.Register_string("verbose_level", "NULL");
  }


#ifdef USE_PARAMETERS_FACTORY
  bool init_param = ParametersFactory::Register("Force.F_Clover_Nf2_Isochemical", append_entry);
#endif
}
//- end

//- parameters class
Parameters_Force_F_Clover_Nf2_Isochemical::Parameters_Force_F_Clover_Nf2_Isochemical() { append_entry(*this); }
//- end

const std::string Force_F_Clover_Nf2_Isochemical::class_name = "Force_F_Clover_Nf2_Isochemical";

//====================================================================
void Force_F_Clover_Nf2_Isochemical::set_parameters(const Parameters& params)
{
  const string str_vlevel = params.get_string("verbose_level");

  m_vl = vout.set_verbose_level(str_vlevel);

  //- fetch and check input parameters
  double        kappa, cSW, mu;
  valarray<int> bc;

  int err = 0;
  err += params.fetch_double("hopping_parameter", kappa);
  err += params.fetch_double("clover_coefficient", cSW);
  err += params.fetch_double("isospin_chemical_potential", mu);
  err += params.fetch_int_vector("boundary_condition", bc);

  if (err) {
    vout.crucial(m_vl, "%s: fetch error, input parameter not found.\n", class_name.c_str());
    abort();
  }


  set_parameters(kappa, cSW, mu, bc);
}


//====================================================================
void Force_F_Clover_Nf2_Isochemical::set_parameters(const double kappa, const double cSW,
                                                    const double mu, const valarray<int> bc)
{
  int Ndim = CommonParameters::Ndim();

  //- print input parameters
  vout.general(m_vl, "Parameters of %s:\n", class_name.c_str());
  vout.general(m_vl, "  kappa = %8.4f\n", kappa);
  vout.general(m_vl, "  cSW   = %8.4f\n", cSW);
  vout.general(m_vl, "  mu    = %8.4f\n", mu);
  for (int imu = 0; imu < Ndim; ++imu) {
    vout.general(m_vl, "  boundary[%d] = %2d\n", imu, bc[imu]);
  }

  //- range check
  // NB. kappa,cSW,mu == 0 is allowed.
  assert(bc.size() == Ndim);

  //- store values
  m_kappa = kappa;
  m_cSW   = cSW;
  m_mu    = mu;

  m_boundary.resize(Ndim);
  for (int dir = 0; dir < Ndim; ++dir) {
    m_boundary[dir] = bc[dir];
  }

  //- propagate parameters
  m_fopr_c->set_parameters(m_kappa, m_cSW, m_mu, m_boundary);

  m_force_w->set_parameters(m_kappa, m_mu, m_boundary);
  m_force_csw->set_parameters(m_kappa, m_cSW, m_boundary);
}


//====================================================================
void Force_F_Clover_Nf2_Isochemical::init(std::string repr)
{
  m_repr = repr;

  m_fopr_c    = new Fopr_Clover_Isochemical(m_repr);
  m_force_w   = new Force_F_Wilson_Nf2_Isochemical(m_repr);
  m_force_csw = new Force_F_CloverTerm(m_repr);

  m_boundary.resize(CommonParameters::Ndim());
  m_Ndim = CommonParameters::Ndim();
}


//====================================================================
void Force_F_Clover_Nf2_Isochemical::tidyup()
{
  delete m_force_csw;
  delete m_force_w;
  delete m_fopr_c;
}


//====================================================================
void Force_F_Clover_Nf2_Isochemical::force_udiv(Field& force_, const Field& eta_)
{
  int Nc   = CommonParameters::Nc();
  int Nvol = CommonParameters::Nvol();
  int Ndim = CommonParameters::Ndim();

  Field_G force(Nvol, Ndim);
  Field_F zeta(Nvol, 1);
  Field_F eta(eta_);

  //  zeta = m_fopr_c->H(eta);
  m_fopr_c->H(zeta, eta);

  set_mode("H");
  force_udiv1_impl(force, zeta, eta);
  force_ = force;

  set_mode("Hdag");
  force_udiv1_impl(force, eta, zeta);
  force_ += force;
}


//====================================================================
void Force_F_Clover_Nf2_Isochemical::force_udiv1(Field& force_, const Field& zeta_, const Field& eta_)
{
  int Nc   = CommonParameters::Nc();
  int Nvol = CommonParameters::Nvol();
  int Ndim = CommonParameters::Ndim();

  Field_G force(Nvol, Ndim);
  Field_F zeta(zeta_);
  Field_F eta(eta_);

  force_udiv1_impl(force, zeta, eta);

  force_ = force;
}


//====================================================================
void Force_F_Clover_Nf2_Isochemical::force_udiv1_impl(Field_G& force, const Field_F& zeta, const Field_F& eta)
{
  int Nc   = CommonParameters::Nc();
  int Nd   = CommonParameters::Nd();
  int Nvol = CommonParameters::Nvol();
  int Ndim = CommonParameters::Ndim();
  // int NinG = 2 * Nc * Nc;

  ShiftField_lex shift;

//  Field force(NinG, Nvol, Ndim);
  force = 0.0;

//  Field force2(NinG, Nvol, Ndim);
  Field_G force2(Nvol, Ndim);

  m_force_w->set_mode(m_mode);
  m_force_w->force_udiv1(force, zeta, eta);

  m_force_csw->force_udiv1(force2, zeta, eta);

  force += force2;
}


//====================================================================
//============================================================END=====
