/*!
        @file    $Id:: force_F_Wilson_Nf2.cpp #$

        @brief

        @author  <Hideo Matsufuru> hideo.matsufuru@kek.jp(matsufuru)
                 $LastChangedBy: aoym $

        @date    $LastChangedDate:: 2014-09-16 12:23:39 #$

        @version $LastChangedRevision: 1144 $
*/

#include "force_F_Wilson_Nf2.h"

#ifdef USE_PARAMETERS_FACTORY
#include "parameters_factory.h"
#endif

using Bridge::vout;

//- parameter entries
namespace {
  void append_entry(Parameters& param)
  {
    param.Register_double("hopping_parameter", 0.0);
    param.Register_int_vector("boundary_condition", std::valarray<int>());

    param.Register_string("verbose_level", "NULL");
  }


#ifdef USE_PARAMETERS_FACTORY
  bool init_param = ParametersFactory::Register("Force.F_Wilson_Nf2", append_entry);
#endif
}
//- end

//- parameters class
Parameters_Force_F_Wilson_Nf2::Parameters_Force_F_Wilson_Nf2() { append_entry(*this); }
//- end

const std::string Force_F_Wilson_Nf2::class_name = "Force_F_Wilson_Nf2";

//====================================================================
void Force_F_Wilson_Nf2::set_parameters(const Parameters& params)
{
  const string str_vlevel = params.get_string("verbose_level");

  m_vl = vout.set_verbose_level(str_vlevel);

  //- fetch and check input parameters
  double        kappa;
  valarray<int> bc;

  int err = 0;
  err += params.fetch_double("hopping_parameter", kappa);
  err += params.fetch_int_vector("boundary_condition", bc);

  if (err) {
    vout.crucial(m_vl, "%s: fetch error, input parameter not found.\n", class_name.c_str());
    abort();
  }


  set_parameters(kappa, bc);
}


//====================================================================
void Force_F_Wilson_Nf2::set_parameters(const double kappa, const valarray<int> bc)
{
  int Ndim = CommonParameters::Ndim();

  //- print input parameters
  vout.general(m_vl, "Parameters of %s:\n", class_name.c_str());
  vout.general(m_vl, "  kappa = %8.4f\n", kappa);
  for (int mu = 0; mu < Ndim; ++mu) {
    vout.general(m_vl, "  boundary[%d] = %2d\n", mu, bc[mu]);
  }

  //- range check
  // NB. kappa == 0 is allowed.
  assert(bc.size() == Ndim);

  //- store values
  m_kappa = kappa;

  m_boundary.resize(Ndim);
  for (int mu = 0; mu < Ndim; ++mu) {
    m_boundary[mu] = bc[mu];
  }

  //- post-process
  m_fopr_w->set_parameters(m_kappa, m_boundary);
}


//====================================================================
void Force_F_Wilson_Nf2::force_udiv(Field& force_, const Field& eta_)
{
  int Nc   = CommonParameters::Nc();
  int Nd   = CommonParameters::Nd();
  int Nvol = CommonParameters::Nvol();
  int Ndim = CommonParameters::Ndim();

  Field_G force1(Nvol, Ndim);
  Field_F zeta(Nvol, 1);
  Field_F eta(eta_);

  m_fopr_w->set_mode("H");
  m_fopr_w->mult(zeta, eta);

  force_udiv1_impl(force1, zeta, eta);
  force_ = force1;

  force_udiv1_impl(force1, eta, zeta);
  force_ += force1;
}


//====================================================================
void Force_F_Wilson_Nf2::force_udiv1(Field& force_, const Field& zeta_, const Field& eta_)
{
  int Nvol = CommonParameters::Nvol();
  int Ndim = CommonParameters::Ndim();

  Field_G force(Nvol, Ndim);
  Field_F zeta(zeta_);
  Field_F eta(eta_);

  force_udiv1_impl(force, zeta, eta);

  force_ = force;
}


//====================================================================
void Force_F_Wilson_Nf2::force_udiv1_impl(Field_G& force, const Field_F& zeta, const Field_F& eta)
{
  int Nc   = CommonParameters::Nc();
  int Nd   = CommonParameters::Nd();
  int Nvol = CommonParameters::Nvol();
  int Ndim = CommonParameters::Ndim();

//  Field_G force(Nvol, Ndim);
  force = 0.0;

//  Field_G force1(Nvol, 1);
  Field_F eta2(Nvol, 1);
  Field_F eta3(Nvol, 1);

  for (int mu = 0; mu < Ndim; ++mu) {
    eta2 = m_fopr_w->mult_gm5p(mu, eta);
    mult_Field_Gd(eta3, 0, *m_U, mu, eta2, 0);
//    m_fopr_w->gm5p(eta3, mu, eta);

//    tensorProd_Field_F(force1, zeta, eta3);
//    force.setpart_ex(mu, force1, 0);
    tensorProd_Field_F(force, mu, zeta, eta3);
  }

  force *= -m_kappa;
}


//====================================================================
//============================================================END=====
