/*!
        @file    $Id: test_Eigensolver_Chebyshev.cpp #$

        @brief

        @author  Hideo Matsufuru <hideo.matsufuru@kek.jp> (matsufuru)
                 $LastChangedBy: aoym $

        @date    $LastChangedDate: 2013-01-22 13:51:53 #$

        @version $LastChangedRevision: 1144 $
*/

#include "test.h"

#include "eigensolver_IRLanczos.h"

#include "gaugeConfig.h"
#include "mat_SU_N.h"
#include "vec_SU_N.h"

#include "fopr_Clover.h"
#include "fopr_Chebyshev.h"

#include "projection.h"
#include "smear.h"
#include "sortField.h"

#include "director_Smear.h"
#include "fopr_Smeared.h"

#include "randomNumbers_Mseries.h"

//====================================================================
//! Test of eigenvalue solver with Chebyshev.

/*!
                                          [28 Dec 2011 H.Matsufuru]
    (Coding history will be recovered from trac.)
    YAML is implemented.                  [14 Nov 2012 Y.Namekawa]
    Selectors are implemented.            [03 Mar 2013 Y.Namekawa]
    (Selectors are replaced with factories by Aoyama-san)
 */

namespace Test_Eigensolver {
  const std::string test_name = "Eigensolver.Solve_Chebyshev";

  //- test-private parameters
  namespace {
    const std::string filename_input  = "test_Eigensolver_Chebyshev.yaml";
    const std::string filename_output = "stdout";

    class Parameters_Test_Eigensolver : public Parameters {
     public:
      Parameters_Test_Eigensolver()
      {
        Register_string("gauge_config_status", "NULL");
        Register_string("gauge_config_type_input", "NULL");
        Register_string("config_filename_input", "NULL");

        Register_string("verbose_level", "NULL");

        Register_double("expected_result", 0.0);
      }
    };
  }

  //- prototype declaration
  int solve_chebyshev(void);

#ifdef USE_TESTMANAGER_AUTOREGISTER
  namespace {
#if defined(USE_GROUP_SU2)
    // Nc=2 is not available.
#else
    static const bool is_registered = TestManager::RegisterTest(
      test_name,
      solve_chebyshev
      );
#endif
  }
#endif

  //====================================================================
  //- Check of eigenvalue solver
  int solve_chebyshev(void)
  {
    // ####  parameter setup  ####
    int Ndim = CommonParameters::Ndim();
    int Nvol = CommonParameters::Nvol();

    Parameters *params_test      = new Parameters_Test_Eigensolver;
    Parameters *params_clover    = ParametersFactory::New("Fopr.Clover");
    Parameters *params_chebyshev = ParametersFactory::New("Fopr.Chebyshev");
    Parameters *params_proj      = ParametersFactory::New("Projection");
    Parameters *params_smear     = ParametersFactory::New("Smear");
    Parameters *params_dr_smear  = ParametersFactory::New("Director_Smear");
    Parameters *params_irlanczos = ParametersFactory::New("Eigensolver");
    Parameters *params_all       = new Parameters;

    params_all->Register_Parameters("Test_Eigensolver", params_test);
    params_all->Register_Parameters("Fopr_Clover", params_clover);
    params_all->Register_Parameters("Fopr_Chebyshev", params_chebyshev);
    params_all->Register_Parameters("Projection", params_proj);
    params_all->Register_Parameters("Smear", params_smear);
    params_all->Register_Parameters("Director_Smear", params_dr_smear);
    params_all->Register_Parameters("Eigensolver_Chebyshev", params_irlanczos);

    ParameterManager_YAML params_manager;
    params_manager.read_params(filename_input, params_all);

    const string str_gconf_status = params_test->get_string("gauge_config_status");
    const string str_gconf_read   = params_test->get_string("gauge_config_type_input");
    const string readfile         = params_test->get_string("config_filename_input");
    const string str_vlevel       = params_test->get_string("verbose_level");

    const bool   do_check        = params_test->is_set("expected_result");
    const double expected_result = do_check ? params_test->get_double("expected_result") : 0.0;

    const string str_gmset_type     = params_clover->get_string("gamma_matrix_type");
    const string str_proj_type      = params_proj->get_string("projection_type");
    const string str_smear_type     = params_smear->get_string("smear_type");
    const string str_sortfield_type = params_irlanczos->get_string("eigensolver_mode");
    const int    Nk          = params_irlanczos->get_int("number_of_wanted_eigenvectors");
    const int    Np          = params_irlanczos->get_int("number_of_working_eigenvectors");
    const int    Niter_eigen = params_irlanczos->get_int("maximum_number_of_iteration");
    const double Enorm_eigen = params_irlanczos->get_double("convergence_criterion_squared");
    const double Vthreshold  = params_irlanczos->get_double("threshold_value");

    Bridge::VerboseLevel vl = vout.set_verbose_level(str_vlevel);

    //- print input parameters
    vout.general(vl, "  gconf_status   = %s\n", str_gconf_status.c_str());
    vout.general(vl, "  gconf_read     = %s\n", str_gconf_read.c_str());
    vout.general(vl, "  readfile       = %s\n", readfile.c_str());
    vout.general(vl, "  vlevel         = %s\n", str_vlevel.c_str());
    vout.general(vl, "  gmset_type     = %s\n", str_gmset_type.c_str());
    vout.general(vl, "  proj_type      = %s\n", str_proj_type.c_str());
    vout.general(vl, "  smear_type     = %s\n", str_smear_type.c_str());
    vout.general(vl, "  sortfield_type = %s\n", str_sortfield_type.c_str());
    vout.general(vl, "\n");

    //- input parameter check
    int err = 0;
    err += ParameterCheck::non_NULL(str_gconf_status);

    if (err) {
      vout.crucial(vl, "%s: input parameters have not been set.\n", test_name.c_str());
      abort();
    }


    // ####  Set up a gauge configuration  ####
    Field_G     *U          = new Field_G(Nvol, Ndim);
    GaugeConfig *gconf_read = new GaugeConfig(str_gconf_read);

    if (str_gconf_status == "Continue") {
      gconf_read->read_file((Field *)U, readfile);
    } else if (str_gconf_status == "Cold_start") {
      U->set_unit();
    } else if (str_gconf_status == "Hot_start") {
      RandomNumbers_Mseries rand(1234567);
      U->set_random(&rand);
    } else {
      vout.crucial(vl, "%s: unsupported gconf status \"%s\".\n", test_name.c_str(), str_gconf_status.c_str());
      abort();
    }

    Projection *proj  = Projection::New(str_proj_type);
    Smear      *smear = Smear::New(str_smear_type, proj);
    smear->set_parameters(*params_smear);

    Director_Smear *dr_smear = new Director_Smear((Smear *)smear);
    dr_smear->set_parameters(*params_dr_smear);


    // ####  object setup  #####
    Fopr_Clover *fopr_c = new Fopr_Clover(str_gmset_type);
    fopr_c->set_parameters(*params_clover);

    Fopr_Smeared *fopr_smear = new Fopr_Smeared(fopr_c, dr_smear);
    Fopr         *fopr       = fopr_smear;
    fopr->set_config(U);
    fopr->set_mode("DdagD");

    Fopr_Chebyshev *fopr_ch = new Fopr_Chebyshev(fopr);
    fopr_ch->set_parameters(*params_chebyshev);
    double Vthreshold_ch = fopr_ch->mult(Vthreshold * Vthreshold);
    vout.general(vl, "Vthreshold_ch = %12.6f\n", Vthreshold_ch);

    //- NB. Low and High must be inversed, because of Chebyshev expansion.
    SortField *sort = SortField::New(str_sortfield_type);

    Eigensolver_IRLanczos *eigen = new Eigensolver_IRLanczos(fopr_ch, sort);
    eigen->set_parameters(Nk, Np, Niter_eigen, Enorm_eigen, Vthreshold_ch);

    Timer *timer = new Timer(test_name);


    // ####  Execution main part  ####
    timer->start();

    int              Nm = Nk + Np;
    valarray<double> TDa(Nm);
    valarray<Field>  vk(Nm);

    Field_F b2;
    int     NFin  = b2.nin();
    int     NFvol = b2.nvol();
    int     NFex  = b2.nex();
    for (int k = 0; k < Nm; ++k) {
      vk[k].reset(NFin, NFvol, NFex);
    }

    int Nsbt  = -1;
    int Nconv = -100;
    eigen->solve(TDa, vk, Nsbt, Nconv, (Field)b2);

    //  vout.general(vl, "  Nsbt  = %d\n", Nsbt);
    //  vout.general(vl, "  Nconv = %d\n", Nconv);

    Field v;
    v.reset(NFin, NFvol, NFex);
    double vv = 0.0;  // superficial initialization

    fopr->set_mode("H");

    for (int i = 0; i < Nsbt + 1; ++i) {
      v = fopr->mult(vk[i]);

      double vnum = vk[i] * v;
      double vden = vk[i] * vk[i];
      double veig = vnum / vden;

      TDa[i] = veig;
      v     -= TDa[i] * vk[i];
      //double vv = v * v;
      vv = v * v;

      vout.general(vl, "Eigenvalues: %4d %20.14f  %10.4e  %10.4e\n", i, TDa[i], vv, vden - 1.0);
    }

    double result = TDa[0];

    timer->report();


    // #####  tidy up  #####
    delete params_test;
    delete params_clover;
    delete params_chebyshev;
    delete params_proj;
    delete params_smear;
    delete params_dr_smear;
    delete params_irlanczos;
    delete params_all;

    delete timer;

    delete U;
    delete gconf_read;

    delete dr_smear;

    delete fopr_c;
    delete fopr_ch;
    delete fopr_smear;

    delete proj;
    delete smear;
    delete sort;

    delete eigen;


    if (do_check) {
      return Test::verify(expected_result, result);
    } else {
      vout.detailed(vl, "check skipped: expected_result not set.\n\n");
      return EXIT_SKIP;
    }
  }
} // namespace Test_Eigensolver
