/*!
        @file    $Id:: test_GradientFlow.cpp #$

        @brief

        @author  Sinya Aoki <saoki@het.ph.tsukuba.ac.jp>

        @date    $LastChangedDate:: 2014-09-16 12:23:39 #$

        @version $LastChangedRevision: 1144 $
*/

#include "test.h"

#include "gaugeConfig.h"

#include "action_G_Plaq.h"
#include "action_G_Rectangle.h"

#include "gradientFlow.h"

#include "randomNumbers_Mseries.h"

//====================================================================
//! Test of gradientFlow.

/*!
                        [08 Jul 2012 S.Aoki]
  (Coding history will be recovered from trac.)
  YAML is implemented.  [14 Nov 2012 Y.Namekawa]
 */

namespace Test_GradientFlow {
  const std::string test_name = "GradientFlow";

  //- test-private parameters
  namespace {
    const std::string filename_input  = "test_GradientFlow.yaml";
    const std::string filename_output = "stdout";

    class Parameters_Test_GradientFlow : public Parameters {
     public:
      Parameters_Test_GradientFlow()
      {
        Register_string("gauge_config_status", "NULL");
        Register_string("gauge_config_type_input", "NULL");
        Register_string("config_filename_input", "NULL");

        Register_string("gauge_config_type_output", "NULL");
        Register_string("config_filename_output", "NULL");

        Register_string("verbose_level", "NULL");

        Register_double("expected_result", 0.0);
      }
    };
  }

  //- prototype declaration
  int update(void);

#ifdef USE_TESTMANAGER_AUTOREGISTER
  namespace {
#if defined(USE_GROUP_SU2)
    // Nc=2 is not available.
#else
    static const bool is_registered = TestManager::RegisterTest(
      test_name,
      update
      );
#endif
  }
#endif

  //====================================================================
  int update(void)
  {
    // ####  parameter setup  ####
    int Nvol = CommonParameters::Nvol();
    int Ndim = CommonParameters::Ndim();

    Parameters *params_test     = new Parameters_Test_GradientFlow;
    Parameters *params_action_G = ParametersFactory::New("Action.G_Plaq");
    Parameters *params_g_flow   = ParametersFactory::New("GradientFlow");
    Parameters *params_all      = new Parameters;

    params_all->Register_Parameters("Test_GradientFlow", params_test);
    params_all->Register_Parameters("Action_G_Plaq", params_action_G);
    params_all->Register_Parameters("GradientFlow", params_g_flow);

    ParameterManager_YAML params_manager;
    params_manager.read_params(filename_input, params_all);

    const string str_gconf_status = params_test->get_string("gauge_config_status");
    const string str_gconf_read   = params_test->get_string("gauge_config_type_input");
    const string readfile         = params_test->get_string("config_filename_input");
    const string str_gconf_write  = params_test->get_string("gauge_config_type_output");
    const string writefile        = params_test->get_string("config_filename_output");
    const string str_vlevel       = params_test->get_string("verbose_level");

    const bool   do_check        = params_test->is_set("expected_result");
    const double expected_result = do_check ? params_test->get_double("expected_result") : 0.0;

    Bridge::VerboseLevel vl = vout.set_verbose_level(str_vlevel);

    //- print input parameters
    vout.general(vl, "  gconf_status = %s\n", str_gconf_status.c_str());
    vout.general(vl, "  gconf_read   = %s\n", str_gconf_read.c_str());
    vout.general(vl, "  readfile     = %s\n", readfile.c_str());
    vout.general(vl, "  gconf_write  = %s\n", str_gconf_write.c_str());
    vout.general(vl, "  writefile    = %s\n", writefile.c_str());
    vout.general(vl, "  vlevel       = %s\n", str_vlevel.c_str());
    vout.general(vl, "\n");

    //- input parameter check
    int err = 0;
    err += ParameterCheck::non_NULL(str_gconf_status);

    if (err) {
      vout.crucial(vl, "%s: Input parameters have not been set.\n", test_name.c_str());
      abort();
    }


    // #### object setup #####
    Field_G     *U          = new Field_G(Nvol, Ndim);
    GaugeConfig *gconf_read = new GaugeConfig(str_gconf_read);

    if (str_gconf_status == "Continue") {
      gconf_read->read_file((Field *)U, readfile);
    } else if (str_gconf_status == "Cold_start") {
      U->set_unit();
    } else if (str_gconf_status == "Hot_start") {
      RandomNumbers_Mseries rand(1234567);
      U->set_random(&rand);
    } else {
      vout.crucial(vl, "%s: unsupported gconf status \"%s\".\n", test_name.c_str(), str_gconf_status.c_str());
      abort();
    }

    Action_G_Plaq *action_G = new Action_G_Plaq;
    action_G->set_parameters(*params_action_G);

    GradientFlow *g_flow = new GradientFlow((Action * )action_G);
    g_flow->set_parameters(*params_g_flow);

    Timer *timer = new Timer(test_name);


    // ####  Execution main part  ####
    timer->start();

    double result = g_flow->evolve(*U); // mod. by (SM)

    timer->report();


    // ####  tydy up  ####
    delete params_test;
    delete params_action_G;
    delete params_g_flow;
    delete params_all;

    delete timer;

    delete U;
    delete gconf_read;

    delete g_flow;
    delete action_G;


    if (do_check) {
      return Test::verify(expected_result, result);
    } else {
      vout.detailed(vl, "check skipped: expected_result not set.\n\n");
      return EXIT_SKIP;
    }
  }
} // namespace Test_GradientFlow
