/*!
        @file    $Id: test_HMC_Clover_Leapfrog_Nf2_eo.cpp #$

        @brief

        @author  <Yusuke Namekawa> namekawa@ccs.tsukuba.ac.jp(namekawa)
                 $LastChangedBy: namekawa $

        @date    $LastChangedDate: 2013-01-22 13:51:53 #$

        @version $LastChangedRevision: 1145 $
*/

#include "test.h"

#include "gaugeConfig.h"

#include "action_G_Plaq.h"
#include "action_G_Rectangle.h"

#include "fopr_Clover.h"
#include "force_F_Clover_Nf2.h"

#include "randomNumbers_Mseries.h"

#include "director_Smear.h"
#include "fopr_Smeared.h"
#include "force_F_Smeared.h"

#include "forceSmear.h"
#include "projection.h"
#include "smear.h"
#include "solver.h"

#include "action_F_Standard_eo.h"
#include "fopr_Clover_eo.h"
#include "fopr_Smeared_eo.h"
#include "fprop_Standard_eo.h"

#include "hmc_Leapfrog.h"

//====================================================================
//! Test of spectroscopy with clover fermion.

/*!
    This test class obtains the quark propagator for clover fermion
    and calculates typical hadron correlators.
    The quantum numbers of hadrons are specified with gamma matrices
    (GammaMatrix class instance) whose set is defined in a subclass
    of GammaMatrixSet class.
                                           [12 Apr 2012 H.Matsufuru]
    (Coding history will be recovered from trac.)
    YAML is implemented.                   [14 Nov 2012 Y.Namekawa]
    Selectors are implemented.             [02 Feb 2013 Y.Namekawa]
    (Selectors are replaced with factories by Aoyama-san)
    Even-odd and smearing are implemented. [03 Mar 2013 Y.Namekawa]
 */

namespace Test_HMC_Clover {
  const std::string test_name = "HMC.Clover.Leapfrog_Nf2_eo";

  //- test-private parameters
  namespace {
    const std::string filename_input  = "test_HMC_Clover_Leapfrog_Nf2.yaml";
    const std::string filename_output = "stdout";

    class Parameters_Test_HMC_Clover : public Parameters {
     public:
      Parameters_Test_HMC_Clover()
      {
        Register_string("gauge_config_status", "NULL");

        Register_string("gauge_config_type_input", "NULL");
        Register_string("config_filename_input", "NULL");

        Register_string("gauge_config_type_output", "NULL");
        Register_string("config_filename_output", "NULL");

        Register_int("trajectory_number", 0);
        Register_int("trajectory_number_step", 0);
        Register_int("save_config_interval", 0);

        Register_string("verbose_level", "NULL");

        Register_double("expected_result", 0.0);
      }
    };
  }

  //- prototype declaration
  int leapfrog_Nf2_eo(void);

#ifdef USE_TESTMANAGER_AUTOREGISTER
  namespace {
#if defined(USE_GROUP_SU2)
    // Nc=2 is not available.
#else
    static const bool is_registered = TestManager::RegisterTest(
      test_name,
      leapfrog_Nf2_eo
      );
#endif
  }
#endif

  //====================================================================
  int leapfrog_Nf2_eo(void)
  {
    // #####  parameter setup  #####
    int Nc   = CommonParameters::Nc();
    int Nvol = CommonParameters::Nvol();
    int Ndim = CommonParameters::Ndim();

    Parameters *params_test      = new Parameters_Test_HMC_Clover;
    Parameters *params_action_G  = ParametersFactory::New("Action.G_Rectangle");
    Parameters *params_clover    = ParametersFactory::New("Fopr.Clover");
    Parameters *params_proj      = ParametersFactory::New("Projection");
    Parameters *params_smear     = ParametersFactory::New("Smear");
    Parameters *params_dr_smear  = ParametersFactory::New("Director_Smear");
    Parameters *params_solver_MD = ParametersFactory::New("Solver");
    Parameters *params_solver_H  = ParametersFactory::New("Solver");
    Parameters *params_hmc       = ParametersFactory::New("HMC.Leapfrog");
    Parameters *params_all       = new Parameters;

    params_all->Register_Parameters("Test_HMC_Clover", params_test);
    params_all->Register_Parameters("Action_G_Rectangle", params_action_G);
    params_all->Register_Parameters("Fopr_Clover", params_clover);
    params_all->Register_Parameters("Projection", params_proj);
    params_all->Register_Parameters("Smear", params_smear);
    params_all->Register_Parameters("Director_Smear", params_dr_smear);
    params_all->Register_Parameters("Solver_MD", params_solver_MD);
    params_all->Register_Parameters("Solver_H", params_solver_H);
    params_all->Register_Parameters("HMC_Leapfrog", params_hmc);

    ParameterManager_YAML params_manager;
    params_manager.read_params(filename_input, params_all);

    const string str_gconf_status = params_test->get_string("gauge_config_status");
    const string str_gconf_read   = params_test->get_string("gauge_config_type_input");
    const string readfile         = params_test->get_string("config_filename_input");
    const string str_gconf_write  = params_test->get_string("gauge_config_type_output");
    const string writefile        = params_test->get_string("config_filename_output");
    int          iconf            = params_test->get_int("trajectory_number");
    const int    Ntraj            = params_test->get_int("trajectory_number_step");
    const int    i_save_conf      = params_test->get_int("save_config_interval");
    const string str_vlevel       = params_test->get_string("verbose_level");

    const bool   do_check        = params_test->is_set("expected_result");
    const double expected_result = do_check ? params_test->get_double("expected_result") : 0.0;

    const string str_gmset_type     = params_clover->get_string("gamma_matrix_type");
    const string str_proj_type      = params_proj->get_string("projection_type");
    const string str_smear_type     = params_smear->get_string("smear_type");
    const string str_solver_MD_type = params_solver_MD->get_string("solver_type");
    const string str_solver_H_type  = params_solver_H->get_string("solver_type");

    Bridge::VerboseLevel vl = vout.set_verbose_level(str_vlevel);

    //- print input parameters
    vout.general(vl, "  gconf_status = %s\n", str_gconf_status.c_str());
    vout.general(vl, "  gconf_read     = %s\n", str_gconf_read.c_str());
    vout.general(vl, "  readfile       = %s\n", readfile.c_str());
    vout.general(vl, "  gconf_write    = %s\n", str_gconf_write.c_str());
    vout.general(vl, "  writefile      = %s\n", writefile.c_str());
    vout.general(vl, "  iconf          = %d\n", iconf);
    vout.general(vl, "  Ntraj          = %d\n", Ntraj);
    vout.general(vl, "  i_save_conf    = %d\n", i_save_conf);
    vout.general(vl, "  vlevel         = %s\n", str_vlevel.c_str());
    vout.general(vl, "  gmset_type     = %s\n", str_gmset_type.c_str());
    vout.general(vl, "  proj_type      = %s\n", str_proj_type.c_str());
    vout.general(vl, "  smear_type     = %s\n", str_smear_type.c_str());
    vout.general(vl, "  solver_MD_type = %s\n", str_solver_MD_type.c_str());
    vout.general(vl, "  solver_H_type  = %s\n", str_solver_H_type.c_str());
    vout.general(vl, "\n");

    //- input parameter check
    int err = 0;
    err += ParameterCheck::non_NULL(str_gconf_status);
    err += ParameterCheck::non_negative(iconf);
    err += ParameterCheck::non_negative(Ntraj);
    err += ParameterCheck::non_negative(i_save_conf);

    if (err) {
      vout.crucial(vl, "%s: Input parameters have not been set.\n", test_name.c_str());
      abort();
    }


    // #####  object setup  #####
    Field_G     *U           = new Field_G(Nvol, Ndim);
    GaugeConfig *gconf_read  = new GaugeConfig(str_gconf_read);
    GaugeConfig *gconf_write = new GaugeConfig(str_gconf_write);

    if (str_gconf_status == "Continue") {
      gconf_read->read_file((Field *)U, readfile);
    } else if (str_gconf_status == "Cold_start") {
      U->set_unit();
    } else if (str_gconf_status == "Hot_start") {
      RandomNumbers_Mseries rand(1234567);
      U->set_random(&rand);
    } else {
      vout.crucial(vl, "%s: unsupported gconf status \"%s\".\n", test_name.c_str(), str_gconf_status.c_str());
      abort();
    }

    Action_G_Rectangle *action_G = new Action_G_Rectangle; // plaq.+rectangle (SA)
    action_G->set_parameters(*params_action_G);

    Fopr_Clover *fopr_c = new Fopr_Clover(str_gmset_type); // define fermion operator (SA)
    fopr_c->set_parameters(*params_clover);

    Force_F_Clover_Nf2 *force_fopr_c = new Force_F_Clover_Nf2(str_gmset_type);
    // define fermion force (SA)
    force_fopr_c->set_parameters(*params_clover);

    // define smearing method (SA)
    Projection *proj  = Projection::New(str_proj_type);
    Smear      *smear = Smear::New(str_smear_type, proj);
    smear->set_parameters(*params_smear);

    // define force smearing method (SA)
    ForceSmear *force_smear = ForceSmear::New(str_smear_type, proj);
    force_smear->set_parameters(*params_smear);

    Director_Smear *dr_smear = new Director_Smear((Smear *)smear);
    dr_smear->set_parameters(*params_dr_smear);

    Fopr_Smeared *fopr_smear = new Fopr_Smeared((Fopr *)fopr_c, dr_smear);
    // define smeared fermion operator (SA)
    Force_F_Smeared *force_fopr_smear
      = new Force_F_Smeared((Force *)force_fopr_c, (ForceSmear *)force_smear, dr_smear);
    // define smeared fermion force (SA)

    //- NB1 fopr_c_eo->set_config is performed in Action_F_Standard_eo.
    //  NB2 fopr_c_eo->set_mode   is performed in fprop.
    Fopr_Clover_eo *fopr_c_eo = new Fopr_Clover_eo(str_gmset_type);
    fopr_c_eo->set_parameters(*params_clover);

    Fopr_Smeared_eo *fopr_smear_eo = new Fopr_Smeared_eo((Fopr_eo *)fopr_c_eo, dr_smear);

    Solver *solver_eo_MD = Solver::New(str_solver_MD_type, fopr_smear_eo);
    solver_eo_MD->set_parameters(*params_solver_MD);
    Fprop *fprop_eo_MD = new Fprop_Standard_eo(solver_eo_MD);

    Solver *solver_eo_H = Solver::New(str_solver_H_type, fopr_smear_eo);
    solver_eo_H->set_parameters(*params_solver_H);
    Fprop *fprop_eo_H = new Fprop_Standard_eo(solver_eo_H);

    Action_F_Standard_eo *action_F
      = new Action_F_Standard_eo((Fopr *)fopr_smear, (Force *)force_fopr_smear,
                                 fprop_eo_MD, fprop_eo_H);
    // define fermion action (SA)


    valarray<Action *> actions(2);
    actions[0] = (Action *)action_G;     // register action[0] (SA)
    actions[1] = (Action *)action_F;     // register action[1] (SA)

    valarray<Director *> directors(1);
    directors[0] = (Director *)dr_smear; // register director[0] (SA)


    //- Random number is initialized with a parameter specified by iconf
    RandomNumbers *rand = new RandomNumbers_Mseries(iconf);

    HMC_Leapfrog hmc(actions, directors, rand); // define hmc_leapfrog (SA)
    hmc.set_parameters(*params_hmc);

    Timer *timer = new Timer(test_name);


    // ####  Execution main part  ####
    timer->start();

    vout.general(vl, "HMC: Ntraj = %d\n", Ntraj);  // a number of trajectory (SA)

    double result = 0.0;
    for (int traj = 0; traj < Ntraj; ++traj) {
      vout.general(vl, "\n");
      vout.general(vl, "traj = %d\n", traj);

      result = hmc.update(*U); // hmc update (SA)

      if ((iconf + traj + 1) % i_save_conf == 0) {
        gconf_write->write_file((Field *)U, writefile);
      }
    }

    timer->report();


    // #####  tidy up  #####
    delete params_test;
    delete params_action_G;
    delete params_clover;
    delete params_proj;
    delete params_smear;
    delete params_dr_smear;
    delete params_solver_MD;
    delete params_solver_H;
    delete params_hmc;
    delete params_all;

    delete timer;

    delete U;
    delete gconf_read;
    delete gconf_write;

    delete rand;

    delete action_G;
    delete action_F;

    delete       fopr_c;
    delete force_fopr_c;

    delete proj;
    delete            smear;
    delete      force_smear;
    delete         dr_smear;
    delete       fopr_smear;
    delete force_fopr_smear;

    delete fopr_c_eo;
    delete fopr_smear_eo;

    delete solver_eo_MD;
    delete  fprop_eo_MD;
    delete solver_eo_H;
    delete  fprop_eo_H;


    if (do_check) {
      return Test::verify(expected_result, result);
    } else {
      vout.detailed(vl, "check skipped: expected_result not set.\n\n");
      return EXIT_SKIP;
    }
  }
} // namespace Test_HMC_Clover
