/*!
        @file    $Id: test_IO_Data_Text.cpp #$

        @brief

        @author  Tatsumi Aoyama <aoym@post.kek.jp> (aoym)
                 $LastChangedBy: aoym $

        @date    $LastChangedDate: 2013-03-21 15:28:34 #$

        @version $LastChangedRevision: 1144 $
*/

#include "test.h"

#include "gaugeConfig.h"

#include "dataIO_Text.h"
#include "bridge_complex.h"

#include "randomNumbers_Mseries.h"

//====================================================================
//! Test of I/O.

/*!
    (Implemented by Aoyama-san.)
    (Coding history will be recovered from trac.)
    YAML is implemented.      [14 Nov 2012 Y.Namekawa]
*/

namespace Test_IO_Data {
  const std::string test_name = "IO.Data.Text";

  //- test-private parameters
  namespace {
    const std::string filename_input  = "test_IO_Data_Text.yaml";
    const std::string filename_output = "stdout";

    class Parameters_Test_IO_GaugeConfig : public Parameters {
     public:
      Parameters_Test_IO_GaugeConfig()
      {
        Register_string("gauge_config_status", "NULL");
        Register_string("gauge_config_type_input", "NULL");
        Register_string("config_filename_input", "NULL");
        Register_string("config_filename_output", "NULL");

        Register_int("data_size", 1024);

        Register_string("verbose_level", "NULL");
        Register_double("expected_result", 0.0);
      }
    };
  }

  //- prototype declaration
  int test_io_data_text(void);

#ifdef USE_TESTMANAGER_AUTOREGISTER
  namespace {
#if defined(USE_GROUP_SU2)
    // Nc=2 is not available.
#else
    static const bool is_registered = TestManager::RegisterTest(
      test_name,
      test_io_data_text
      );
#endif
  }
#endif

  //====================================================================
  int test_io_data_text(void)
  {
    // ####  parameter setup  ####
    int          Ndim = CommonParameters::Ndim();
    int          Nvol = CommonParameters::Nvol();
    const double tiny = CommonParameters::epsilon_criterion();

    Parameters *params_test = new Parameters_Test_IO_GaugeConfig;
    Parameters *params_all  = new Parameters;

    params_all->Register_Parameters("Test_IO_Data", params_test);

    ParameterManager_YAML params_manager;
    params_manager.read_params(filename_input, params_all);

    const string str_gconf_status = params_test->get_string("gauge_config_status");
    const string str_gconf_read   = params_test->get_string("gauge_config_type_input");
    const string readfile         = params_test->get_string("config_filename_input");
    const string testfile         = params_test->get_string("config_filename_output");
    const int    data_size        = params_test->get_int("data_size");
    const string str_vlevel       = params_test->get_string("verbose_level");

    const bool   do_check        = params_test->is_set("expected_result");
    const double expected_result = do_check ? params_test->get_double("expected_result") : 0.0;

    Bridge::VerboseLevel vl = vout.set_verbose_level(str_vlevel);

    //- print input parameters
    vout.general(vl, "  gconf_status = %s\n", str_gconf_status.c_str());
    vout.general(vl, "  gconf_read   = %s\n", str_gconf_read.c_str());
    vout.general(vl, "  readfile     = %s\n", readfile.c_str());
    vout.general(vl, "  testfile     = %s\n", testfile.c_str());
    vout.general(vl, "  data_size    = %d\n", data_size);
    vout.general(vl, "  vlevel       = %s\n", str_vlevel.c_str());
    vout.general(vl, "\n");

    //- input parameter check
    int err = 0;
    err += ParameterCheck::non_NULL(str_gconf_status);
    err += ParameterCheck::non_NULL(testfile);

    if (err) {
      vout.crucial(vl, "%s: Input parameters have not been set.\n", test_name.c_str());
      abort();
    }


    // ####  Set up a gauge configuration  ####
    Field_G     *U          = new Field_G(Nvol, Ndim);
    GaugeConfig *gconf_read = new GaugeConfig(str_gconf_read);

    if (str_gconf_status == "Continue") {
      gconf_read->read_file((Field *)U, readfile);
    } else if (str_gconf_status == "Cold_start") {
      U->set_unit();
    } else if (str_gconf_status == "Hot_start") {
      RandomNumbers_Mseries rand(1234567);
      U->set_random(&rand);
    } else {
      vout.crucial(vl, "%s: unsupported gconf status \"%s\".\n", test_name.c_str(), str_gconf_status.c_str());
      abort();
    }


    // #####  object setup  #####
    DataIO *data_io = new DataIO_Text;
    Timer  *timer   = new Timer(test_name);


    // ####  Execution main part  ####
    timer->start();

    err = 0;

    //- first write to file.
    std::valarray<double> array(data_size);
    double                *p = U->ptr(0);

    for (size_t i = 0; i < data_size; ++i) {
      array[i] = *p++;
    }

    data_io->write_file(array, testfile, false);


    //- write again with complex data.
    std::valarray<dcomplex> arrayc(data_size);
    p = U->ptr(0);

    for (size_t i = 0; i < data_size; ++i) {
      double x = *p++;
      double y = *p++;

      arrayc[i] = cmplx(x, y);
    }

    data_io->write_file(arrayc, testfile, true);


    //- then read back from file.
    std::valarray<double> array2(data_size);

    data_io->read_file(array2, testfile);


    //- check with reference config.
    for (size_t i = 0; i < data_size; ++i) {
      if (abs(array[i] - array2[i]) > tiny) ++err;
    }

    vout.general(vl, "Test_IO_Data:\t%s\n", (err == 0) ? "ok" : "failed");


    int result = err;

    timer->report();


    // ####  tydy up  ####
    delete params_test;
    delete params_all;

    delete timer;

    delete U;
    delete gconf_read;
    delete data_io;


    if (do_check) {
      return Test::verify(expected_result, result);
    } else {
      vout.detailed(vl, "check skipped: expected_result not set.\n\n");
      return EXIT_SKIP;
    }
  }
} // namespace Test_IO_Data
