/*!
        @file    $Id:: test_Mult_Wilson.cpp #$

        @brief

        @author  Yusuke Namekawa <namekawa@ccs.tsukuba.ac.jp> (namekawa)
                 $LastChangedBy: namekawa $

        @date    $LastChangedDate:: 2014-09-16 18:27:22 #$

        @version $LastChangedRevision: 1145 $
*/

#include "test.h"

#include "gaugeConfig.h"
#include "randomNumbers_Mseries.h"

#include "fopr_Wilson.h"
#include "gammaMatrixSet.h"

//- profiler of fx10
//#include "fj_tool/fapp.h"

//====================================================================
//! Test of Mult with Wilson fermion.

/*!
                                          [06 Jun 2014 Y.Namekawa]
 */

namespace Test_Mult_Wilson {
  const std::string test_name = "Mult.Wilson";

  //- test-private parameters
  namespace {
    const std::string filename_input  = "test_Mult_Wilson.yaml";
    const std::string filename_output = "stdout";

    class Parameters_Test_Mult_Wilson : public Parameters {
     public:
      Parameters_Test_Mult_Wilson()
      {
        Register_string("gauge_config_status", "NULL");
        Register_string("gauge_config_type_input", "NULL");
        Register_string("config_filename_input", "NULL");

        Register_int("number_of_mult", 0);

        Register_string("verbose_level", "NULL");

        Register_double("expected_result", 0.0);
      }
    };
  }

  //- prototype declaration
  int mult(void);

#ifdef USE_TESTMANAGER_AUTOREGISTER
  namespace {
#if defined(USE_GROUP_SU2)
    // Nc=2 is not available.
#else
    static const bool is_registered = TestManager::RegisterTest(
      test_name,
      mult
      );
#endif
  }
#endif

  //====================================================================
  int mult(void)
  {
    // ####  parameter setup  ####
    int Nc   = CommonParameters::Nc();
    int Nd   = CommonParameters::Nd();
    int Ndim = CommonParameters::Ndim();
    int Nvol = CommonParameters::Nvol();

    int Lvol    = CommonParameters::Lvol();
    int NPE     = CommonParameters::NPE();
    int Nthread = ThreadManager_OpenMP::get_num_threads_available();

    Parameters *params_test   = new Parameters_Test_Mult_Wilson;
    Parameters *params_wilson = ParametersFactory::New("Fopr.Wilson");
    Parameters *params_all    = new Parameters;

    params_all->Register_Parameters("Test_Mult_Wilson", params_test);
    params_all->Register_Parameters("Fopr_Wilson", params_wilson);

    ParameterManager_YAML params_manager;
    params_manager.read_params(filename_input, params_all);

    const string str_gconf_status = params_test->get_string("gauge_config_status");
    const string str_gconf_read   = params_test->get_string("gauge_config_type_input");
    const string readfile         = params_test->get_string("config_filename_input");
    const int    Nmult            = params_test->get_int("number_of_mult");
    const string str_vlevel       = params_test->get_string("verbose_level");

    const bool   do_check        = params_test->is_set("expected_result");
    const double expected_result = do_check ? params_test->get_double("expected_result") : 0.0;

    const string str_gmset_type = params_wilson->get_string("gamma_matrix_type");

    Bridge::VerboseLevel vl = vout.set_verbose_level(str_vlevel);

    //- print input parameters
    vout.general(vl, "  gconf_status = %s\n", str_gconf_status.c_str());
    vout.general(vl, "  gconf_read   = %s\n", str_gconf_read.c_str());
    vout.general(vl, "  readfile     = %s\n", readfile.c_str());
    vout.general(vl, "  Nmult        = %d\n", Nmult);
    vout.general(vl, "  vlevel       = %s\n", str_vlevel.c_str());
    vout.general(vl, "  gmset_type   = %s\n", str_gmset_type.c_str());

    //- input parameter check
    int err = 0;
    err += ParameterCheck::non_NULL(str_gconf_status);

    if (err) {
      vout.crucial(vl, "%s: Input parameters have not been set.\n", test_name.c_str());
      abort();
    }


    // ####  Set up a gauge configuration  ####
    Field_G     *U          = new Field_G(Nvol, Ndim);
    GaugeConfig *gconf_read = new GaugeConfig(str_gconf_read);

    if (str_gconf_status == "Continue") {
      gconf_read->read_file((Field *)U, readfile);
    } else if (str_gconf_status == "Cold_start") {
      U->set_unit();
    } else if (str_gconf_status == "Hot_start") {
      RandomNumbers_Mseries rand(1234567);
      U->set_random(&rand);
    } else {
      vout.crucial(vl, "%s: unsupported gconf status \"%s\".\n", test_name.c_str(), str_gconf_status.c_str());
      abort();
    }


    // ####  object setup  #####
    GammaMatrixSet *gmset = GammaMatrixSet::New(str_gmset_type);

    Fopr_Wilson *fopr_w = new Fopr_Wilson(str_gmset_type);
    fopr_w->set_parameters(*params_wilson);
    fopr_w->set_config(U);
    fopr_w->set_mode("D");

    Timer *timer = new Timer(test_name);

    vout.general(vl, "\n");


    // ####  Execution main part  ####
    Field_F b, y;
    b = 1.0;

    double result = 0.0;


    timer->start();
    //- profiler of fx10 starts
    // fapp_start("Mult.Wilson",1,1);

#pragma omp parallel
    {
      for (int i = 0; i < Nmult; ++i) {
        fopr_w->mult(y, b);
      }
      result = y.norm();
    }

    //- profiler of fx10 ends
    // fapp_stop("Mult.Wilson",1,1);
    timer->stop();
    double elapse_sec = timer->elapsed_sec();


    //- Flops counting
    double gflo_mult   = fopr_w->flop_count() / 1.0e+9;
    double gflops_mult = gflo_mult * Nmult / (elapse_sec * NPE * Nthread);

    vout.general(vl, "%s: %12.4e GFlops / core\n", test_name.c_str(), gflops_mult);
    vout.general(vl, "\n");


    // ####  tidy up  ####
    delete params_test;
    delete params_wilson;
    delete params_all;

    delete timer;

    delete gconf_read;
    delete U;

    delete fopr_w;
    delete gmset;


    if (do_check) {
      return Test::verify(expected_result, result);
    } else {
      vout.detailed(vl, "check skipped: expected_result not set.\n\n");
      return EXIT_SKIP;
    }
  }
} // namespace Test_Mult_Wilson
