/*!
        @file    $Id: test_QuarkNumberSusceptibility_Clover_Isochemical.cpp #$

        @brief

        @author  Hideo Matsufuru <hideo.matsufuru@kek.jp> (matsufuru)
                 $LastChangedBy: namekawa $

        @date    $LastChangedDate: 2013-01-22 13:51:53 #$

        @version $LastChangedRevision: 1145 $
*/

#include "test.h"

#include "gaugeConfig.h"

#include "fopr_Clover_Isochemical.h"
#include "fprop_Standard_lex.h"
#include "noiseVector_Z2.h"
#include "quarkNumberSusceptibility_Wilson.h"

#include "randomNumbers_Mseries.h"

#include "director_Smear.h"
#include "fopr_Smeared.h"

#include "projection.h"
#include "smear.h"
#include "solver.h"

//====================================================================
//! Quark number susceptibility for the Wilson-type fermion.

/*!
    This class tests measurements of the traces which is used to determine
    the quark number susceptibility.
                                [02 Sep 2012 H.Matsufuru]
    (Coding history will be recovered from trac.)
    YAML is implemented.        [14 Nov 2012 Y.Namekawa]
    Selectors are implemented.  [03 Mar 2013 Y.Namekawa]
    (Selectors are replaced with factories by Aoyama-san)
 */

namespace Test_QuarkNumSuscept {
  const std::string test_name = "QuarkNumberSusceptibility.Clover_Isochemical";

  // test-private parameters
  namespace {
    const std::string filename_input  = "test_QuarkNumberSusceptibility_Clover_Isochemical.yaml";
    const std::string filename_output = "stdout";

    class Parameters_Test_QuarkNumSuscept_Clover_Isochemical : public Parameters {
     public:
      Parameters_Test_QuarkNumSuscept_Clover_Isochemical()
      {
        Register_string("gauge_config_status", "NULL");
        Register_string("gauge_config_type_input", "NULL");
        Register_string("config_filename_input", "NULL");

        Register_int("seed_for_noise", 0);
        Register_int("number_of_noises", 0);

        Register_string("verbose_level", "NULL");

        Register_double("expected_result", 0.0);
      }
    };
  }

  // prototype declaration
  int quark_num_suscept(void);

#ifdef USE_TESTMANAGER_AUTOREGISTER
  namespace {
#if defined(USE_GROUP_SU2)
    // Nc=2 is not available.
#else
    static const bool is_registered = TestManager::RegisterTest(
      test_name,
      quark_num_suscept
      );
#endif
  }
#endif

  //====================================================================
  int quark_num_suscept(void)
  {
    // #####  parameter setup  #####
    int Nc   = CommonParameters::Nc();
    int Nvol = CommonParameters::Nvol();
    int Ndim = CommonParameters::Ndim();

    Parameters *params_test     = new Parameters_Test_QuarkNumSuscept_Clover_Isochemical;
    Parameters *params_clover   = ParametersFactory::New("Fopr.Clover_Isochemical");
    Parameters *params_proj     = ParametersFactory::New("Projection");
    Parameters *params_smear    = ParametersFactory::New("Smear");
    Parameters *params_dr_smear = ParametersFactory::New("Director_Smear");
    Parameters *params_solver   = ParametersFactory::New("Solver");
    Parameters *params_all      = new Parameters;

    params_all->Register_Parameters("Test_QuarkNumSuscept_Clover_Isochemical", params_test);
    params_all->Register_Parameters("Fopr_Clover_Isochemical", params_clover);
    params_all->Register_Parameters("Projection", params_proj);
    params_all->Register_Parameters("Smear", params_smear);
    params_all->Register_Parameters("Director_Smear", params_dr_smear);
    params_all->Register_Parameters("Solver", params_solver);

    ParameterManager_YAML params_manager;
    params_manager.read_params(filename_input, params_all);

    const string str_gconf_status = params_test->get_string("gauge_config_status");
    const string str_gconf_read   = params_test->get_string("gauge_config_type_input");
    const string readfile         = params_test->get_string("config_filename_input");
    int          i_seed_noise     = params_test->get_int("seed_for_noise");
    const string str_vlevel       = params_test->get_string("verbose_level");

    const bool   do_check        = params_test->is_set("expected_result");
    const double expected_result = do_check ? params_test->get_double("expected_result") : 0.0;

    const string str_gmset_type  = params_clover->get_string("gamma_matrix_type");
    const string str_proj_type   = params_proj->get_string("projection_type");
    const string str_smear_type  = params_smear->get_string("smear_type");
    const string str_solver_type = params_solver->get_string("solver_type");

    Bridge::VerboseLevel vl = vout.set_verbose_level(str_vlevel);

    //- print input parameters
    vout.general(vl, "  gconf_status = %s\n", str_gconf_status.c_str());
    vout.general(vl, "  gconf_read   = %s\n", str_gconf_read.c_str());
    vout.general(vl, "  readfile     = %s\n", readfile.c_str());
    vout.general(vl, "  i_seed_noise = %d\n", i_seed_noise);
    vout.general(vl, "  vlevel       = %s\n", str_vlevel.c_str());
    vout.general(vl, "  gmset_type   = %s\n", str_gmset_type.c_str());
    vout.general(vl, "  proj_type    = %s\n", str_proj_type.c_str());
    vout.general(vl, "  smear_type   = %s\n", str_smear_type.c_str());
    vout.general(vl, "  solver_type  = %s\n", str_solver_type.c_str());

    //- input parameter check
    int err = 0;
    err += ParameterCheck::non_NULL(str_gconf_status);
    err += ParameterCheck::non_zero(i_seed_noise);

    if (err) {
      vout.crucial(vl, "%s: Input parameters have not been set.\n", test_name.c_str());
      abort();
    }


    // #####  Set up a gauge configuration  ####
    Field_G     *U          = new Field_G(Nvol, Ndim);
    GaugeConfig *gconf_read = new GaugeConfig(str_gconf_read);

    if (str_gconf_status == "Continue") {
      gconf_read->read_file((Field *)U, readfile);
    } else if (str_gconf_status == "Cold_start") {
      U->set_unit();
    } else if (str_gconf_status == "Hot_start") {
      RandomNumbers_Mseries rand(1234567);
      U->set_random(&rand);
    } else {
      vout.crucial(vl, "%s: unsupported gconf status \"%s\".\n", test_name.c_str(), str_gconf_status.c_str());
      abort();
    }


    // #####  object setup  #####
    Fopr_Clover_Isochemical *fopr_c = new Fopr_Clover_Isochemical(str_gmset_type);
    fopr_c->set_parameters(*params_clover);

    Projection *proj  = Projection::New(str_proj_type);
    Smear      *smear = Smear::New(str_smear_type, proj);
    smear->set_parameters(*params_smear);
    Director_Smear *dr_smear = new Director_Smear((Smear *)smear);
    dr_smear->set_parameters(*params_dr_smear);

    Fopr_Smeared *fopr_smear = new Fopr_Smeared((Fopr *)fopr_c, dr_smear);
    fopr_smear->set_config(U);


    //- Random number is initialized with a parameter specified by iseed
    RandomNumbers *rand = new RandomNumbers_Mseries(i_seed_noise);
    NoiseVector   *nv   = new NoiseVector_Z2(rand);

    Solver *solver = Solver::New(str_solver_type, fopr_smear);
    solver->set_parameters(*params_solver);
    Fprop *fprop_lex = new Fprop_Standard_lex(solver);

    QuarkNumberSusceptibility_Wilson *quark_suscept
      = new QuarkNumberSusceptibility_Wilson(fopr_smear, fprop_lex, nv);
    quark_suscept->set_parameters(*params_test);

    Timer *timer = new Timer(test_name);


    // ####  Execution main part  ####
    timer->start();

    double result = 0.0;
    result = quark_suscept->measure();

    timer->report();


    // #####  tidy up  #####
    delete params_test;
    delete params_clover;
    delete params_proj;
    delete params_smear;
    delete params_dr_smear;
    delete params_solver;
    delete params_all;

    delete U;
    delete gconf_read;

    delete fopr_c;

    delete proj;
    delete smear;
    delete dr_smear;
    delete fopr_smear;

    delete rand;
    delete nv;
    delete solver;
    delete fprop_lex;
    delete quark_suscept;

    delete timer;


    if (do_check) {
      return Test::verify(expected_result, result);
    } else {
      vout.detailed(vl, "check skipped: expected_result not set.\n\n");
      return EXIT_SKIP;
    }
  }
} // namespace Test_QuarkNumSuscept
