/*!
        @file    $Id:: test_Spectrum_Wilson_2ptFunction.cpp #$

        @brief

        @author  UEDA, Satoru <sueda@post.kek.jp> (sueda)
                 $LastChangedBy: aoym $

        @date    $LastChangedDate:: 2014-09-16 12:23:39 #$

        @version $LastChangedRevision: 1144 $
*/

#include "test.h"

#include "gaugeConfig.h"
#include "staples.h"

#include "randomNumbers_Mseries.h"

#include "fopr_Wilson.h"
#include "fprop_Standard_lex.h"

#include "gammaMatrixSet.h"
#include "gaugeFixing.h"
#include "solver.h"
#include "source.h"

#include "corr2pt_4spinor.h"

//====================================================================
//! Test of Wilson fermion solver.

/*!
                                          [28 Dec 2011 H.Matsufuru]
    (Coding history will be recovered from trac.)
    YAML is implemented.                  [14 Nov 2012 Y.Namekawa]
    Selectors are implemented.            [03 Mar 2013 Y.Namekawa]
    (Selectors are replaced with factories by Aoyama-san)
 */

namespace Test_Spectrum_Wilson {
  const std::string test_name = "Spectrum.Wilson.Hadron2ptFunction";

  //- test-private parameters
  namespace {
    const std::string filename_input  = "test_Spectrum_Wilson_Hadron2ptFunction.yaml";
    const std::string filename_output = "stdout";

    class Parameters_Test_Spectrum_Wilson : public Parameters {
     public:
      Parameters_Test_Spectrum_Wilson()
      {
        Register_string("gauge_config_status", "NULL");
        Register_string("gauge_config_type_input", "NULL");
        Register_string("config_filename_input", "NULL");

        Register_string("verbose_level", "NULL");

        Register_double("expected_result", 0.0);
      }
    };
  }

  //- prototype declaration
  int hadron_2ptFunction(void);

#ifdef USE_TESTMANAGER_AUTOREGISTER
  namespace {
#if defined(USE_GROUP_SU2)
    // Nc=2 is not available.
#else
    static const bool is_registered = TestManager::RegisterTest(
      test_name,
      hadron_2ptFunction
      );
#endif
  }
#endif

  //====================================================================
  int hadron_2ptFunction(void)
  {
    // ####  parameter setup  ####
    int Nc   = CommonParameters::Nc();
    int Nd   = CommonParameters::Nd();
    int Ndim = CommonParameters::Ndim();
    int Nvol = CommonParameters::Nvol();

    Parameters *params_test   = new Parameters_Test_Spectrum_Wilson;
    Parameters *params_gfix   = ParametersFactory::New("GaugeFixing");
    Parameters *params_wilson = ParametersFactory::New("Fopr.Wilson");
    Parameters *params_solver = ParametersFactory::New("Solver");
    Parameters *params_source = ParametersFactory::New("Source");
    Parameters *params_all    = new Parameters;

    params_all->Register_Parameters("Test_Spectrum_Wilson", params_test);
    params_all->Register_Parameters("GaugeFixing", params_gfix);
    params_all->Register_Parameters("Fopr_Wilson", params_wilson);
    params_all->Register_Parameters("Solver", params_solver);
    params_all->Register_Parameters("Source", params_source);

    ParameterManager_YAML params_manager;
    params_manager.read_params(filename_input, params_all);

    const string str_gconf_status = params_test->get_string("gauge_config_status");
    const string str_gconf_read   = params_test->get_string("gauge_config_type_input");
    const string readfile         = params_test->get_string("config_filename_input");
    const string str_vlevel       = params_test->get_string("verbose_level");

    const bool   do_check        = params_test->is_set("expected_result");
    const double expected_result = do_check ? params_test->get_double("expected_result") : 0.0;

    const string str_gfix_type   = params_gfix->get_string("gauge_fixing_type");
    const string str_gmset_type  = params_wilson->get_string("gamma_matrix_type");
    const string str_solver_type = params_solver->get_string("solver_type");
    const string str_source_type = params_source->get_string("source_type");

    Bridge::VerboseLevel vl = vout.set_verbose_level(str_vlevel);

    //- print input parameters
    vout.general(vl, "  gconf_status = %s\n", str_gconf_status.c_str());
    vout.general(vl, "  gconf_read   = %s\n", str_gconf_read.c_str());
    vout.general(vl, "  readfile     = %s\n", readfile.c_str());
    vout.general(vl, "  vlevel       = %s\n", str_vlevel.c_str());
    vout.general(vl, "  gfix_type    = %s\n", str_gfix_type.c_str());
    vout.general(vl, "  gmset_type   = %s\n", str_gmset_type.c_str());
    vout.general(vl, "  solver_type  = %s\n", str_solver_type.c_str());
    vout.general(vl, "  source_type  = %s\n", str_source_type.c_str());

    //- input parameter check
    int err = 0;
    err += ParameterCheck::non_NULL(str_gconf_status);

    if (err) {
      vout.crucial(vl, "%s: Input parameters have not been set.\n", test_name.c_str());
      abort();
    }


    // ####  Set up a gauge configuration  ####
    Field_G     *U          = new Field_G(Nvol, Ndim);
    GaugeConfig *gconf_read = new GaugeConfig(str_gconf_read);

    if (str_gconf_status == "Continue") {
      gconf_read->read_file((Field *)U, readfile);
    } else if (str_gconf_status == "Cold_start") {
      U->set_unit();
    } else if (str_gconf_status == "Hot_start") {
      RandomNumbers_Mseries rand(1234567);
      U->set_random(&rand);
    } else {
      vout.crucial(vl, "%s: unsupported gconf status \"%s\".\n", test_name.c_str(), str_gconf_status.c_str());
      abort();
    }


    // ####  Gauge fixing  ####
    Staples *staple = new Staples;
    Field_G *Ufix   = new Field_G(Nvol, Ndim);

    int           ndelay = 1000;
    RandomNumbers *rand  = new RandomNumbers_Mseries(ndelay);

    GaugeFixing *gfix = GaugeFixing::New(str_gfix_type, rand);
    gfix->set_parameters(*params_gfix);

    double plaq = staple->plaquette(*U);
    vout.general(vl, "plaq(original) = %18.14f\n", plaq);

    gfix->fix(*Ufix, *U);

    double plaq2 = staple->plaquette(*Ufix);
    vout.general(vl, "plaq(fixed)    = %18.14f\n", plaq2);
    vout.general(vl, "plaq(diff)     = %18.10e\n", plaq - plaq2);


    // ####  object setup  #####
    GammaMatrixSet *gmset = GammaMatrixSet::New(str_gmset_type);

    Fopr_Wilson *fopr_w = new Fopr_Wilson(str_gmset_type);
    fopr_w->set_parameters(*params_wilson);
    fopr_w->set_config(Ufix);

    Solver *solver = Solver::New(str_solver_type, fopr_w);
    solver->set_parameters(*params_solver);

    Fprop *fprop_lex = new Fprop_Standard_lex(solver);

    Source *source = Source::New(str_source_type);
    source->set_parameters(*params_source);

    Timer *timer = new Timer(test_name);

    // ####  flops measurement  ####
    int     Nrepeat     = 500;
    int     Lvol        = CommonParameters::Lvol();
    Timer   *timer_meas = new Timer();
    Field_F v1, v2;
    source->set(v1, 0);

    fopr_w->set_mode("D");
    double fop1          = fopr_w->flop_count();
    double fop1_per_site = fop1 / Lvol;
    double fop           = fop1 * double(2 * Nrepeat);

    timer_meas->start();
#pragma omp parallel
    {
      for (int i = 0; i < Nrepeat; ++i) {
        fopr_w->mult(v2, v1);
        fopr_w->mult_dag(v1, v2);
      }
    }
    timer_meas->stop();
    double elapsed_time = timer_meas->elapsed_sec();
    double flops        = fop / elapsed_time;
    double gflops       = flops * 1.0e-9;

    vout.general(vl, "\n");
    vout.general(vl, "Performance of Wilson(lex) fermion mult:\n");
    vout.general(vl, "  Flop(Fop)    = %18.9e\n", fop1);
    vout.general(vl, "  Flop/site    = %18.9e\n", fop1_per_site);
    vout.general(vl, "  Flop(total)  = %18.9e\n", fop);
    vout.general(vl, "  Elapsed time = %15.6f sec\n", elapsed_time);
    vout.general(vl, "  Performance  = %12.3f GFlops\n", gflops);
    delete timer_meas;

    // ####  Execution main part  ####
    timer->start();

    std::valarray<Field_F> sq(Nc * Nd);
    for (int i = 0; i < Nc * Nd; ++i) {
      sq[i] = 0.0;
    }

    Field_F b;
    b = 0.0;

    int    Nconv;
    double diff;

    vout.general(vl, "\n");
    vout.general(vl, "Solving quark propagator:\n");
    vout.general(vl, "  color spin   Nconv      diff           diff2\n");

    for (int ispin = 0; ispin < Nd; ++ispin) {
      for (int icolor = 0; icolor < Nc; ++icolor) {
        int idx = icolor + Nc * ispin;
        source->set(b, idx);

        fprop_lex->invert_D(sq[idx], b, Nconv, diff);

        Field_F y(b);
        fopr_w->set_mode("D");
        y -= (Field_F)fopr_w->mult(sq[idx]);
        double diff2 = y.norm();

        vout.general(vl, "   %2d   %2d   %6d   %12.4e   %12.4e\n",
                     icolor, ispin, Nconv, diff, diff2);
      }
    }

    //- meson correlators
    vout.general(vl, "\n");
    vout.general(vl, "2-point correlator:\n");

    Corr2pt_4spinor corr(gmset);
    double          result = corr.meson_all(sq, sq);

    timer->report();


    // ####  tidy up  ####
    delete params_test;
    delete params_gfix;
    delete params_wilson;
    delete params_solver;
    delete params_source;
    delete params_all;

    delete timer;

    delete gconf_read;
    delete U;
    delete Ufix;

    delete rand;
    delete staple;

    delete fopr_w;
    delete fprop_lex;

    delete gfix;
    delete gmset;
    delete solver;
    delete source;


    if (do_check) {
      return Test::verify(expected_result, result);
    } else {
      vout.detailed(vl, "check skipped: expected_result not set.\n\n");
      return EXIT_SKIP;
    }
  }
} // namespace Test_Spectrum_Wilson
