/*!
        @file    $Id: test_TopologicalCharge.cpp #$

        @brief

        @author  Yusuke Namekawa <namekawa@ccs.tsukuba.ac.jp> (namekawa)
                 $LastChangedBy: aoym $

        @date    $LastChangedDate: 2013-01-22 13:51:53 #$

        @version $LastChangedRevision: 1144 $
*/

#include "test.h"

#include "gaugeConfig.h"

#include "topologicalCharge.h"

#include "projection.h"
#include "smear.h"

#include "randomNumbers_Mseries.h"

//====================================================================
//! Test of Topological Charge measurement.

/*!
    This class measures the topological charge
    defined by the clover leaf on the lattice.
                          [01 Jan 2014 Y.Namekawa]
 */

namespace Test_TopologicalCharge {
  const std::string test_name = "TopologicalCharge";

  //- test-private parameters
  namespace {
    const std::string filename_input  = "test_TopologicalCharge.yaml";
    const std::string filename_output = "stdout";

    class Parameters_Test_TopologicalCharge : public Parameters {
     public:
      Parameters_Test_TopologicalCharge()
      {
        Register_string("gauge_config_status", "NULL");
        Register_string("gauge_config_type_input", "NULL");
        Register_string("config_filename_input", "NULL");

        Register_int("number_of_max_smearing", 0);
        Register_int("number_of_smearing_step", 0);

        Register_string("verbose_level", "NULL");

        Register_double("expected_result", 0.0);
      }
    };
  }

  //- prototype declaration
  int topological_charge(void);

#ifdef USE_TESTMANAGER_AUTOREGISTER
  namespace {
#if defined(USE_GROUP_SU2)
    // Nc=2 is not available.
#else
    static const bool is_registered = TestManager::RegisterTest(
      test_name,
      topological_charge
      );
#endif
  }
#endif

  //====================================================================
  int topological_charge(void)
  {
    // ####  parameter setup  ####
    int Nvol = CommonParameters::Nvol();
    int Ndim = CommonParameters::Ndim();

    Parameters *params_test  = new Parameters_Test_TopologicalCharge;
    Parameters *params_topo  = ParametersFactory::New("TopologicalCharge");
    Parameters *params_proj  = ParametersFactory::New("Projection");
    Parameters *params_smear = ParametersFactory::New("Smear");
    Parameters *params_all   = new Parameters;

    params_all->Register_Parameters("Test_TopologicalCharge", params_test);
    params_all->Register_Parameters("TopologicalCharge", params_topo);
    params_all->Register_Parameters("Projection", params_proj);
    params_all->Register_Parameters("Smear", params_smear);

    ParameterManager_YAML params_manager;
    params_manager.read_params(filename_input, params_all);

    const string str_gconf_status = params_test->get_string("gauge_config_status");
    const string str_gconf_read   = params_test->get_string("gauge_config_type_input");
    const string readfile         = params_test->get_string("config_filename_input");
    const int    Nsmear           = params_test->get_int("number_of_max_smearing");
    const int    Nmeas            = params_test->get_int("number_of_smearing_step");
    const string str_vlevel       = params_test->get_string("verbose_level");

    const bool   do_check        = params_test->is_set("expected_result");
    const double expected_result = do_check ? params_test->get_double("expected_result") : 0.0;

    const string str_proj_type  = params_proj->get_string("projection_type");
    const string str_smear_type = params_smear->get_string("smear_type");

    Bridge::VerboseLevel vl = vout.set_verbose_level(str_vlevel);

    //- print input parameters
    vout.general(vl, "  gconf_status = %s\n", str_gconf_status.c_str());
    vout.general(vl, "  gconf_read   = %s\n", str_gconf_read.c_str());
    vout.general(vl, "  readfile     = %s\n", readfile.c_str());
    vout.general(vl, "  Nsmear       = %d\n", Nsmear);
    vout.general(vl, "  Nmeas        = %d\n", Nmeas);
    vout.general(vl, "  vlevel       = %s\n", str_vlevel.c_str());
    vout.general(vl, "  proj_type    = %s\n", str_proj_type.c_str());
    vout.general(vl, "  smear_type   = %s\n", str_smear_type.c_str());
    vout.general(vl, "\n");

    //- input parameter check
    int err = 0;
    err += ParameterCheck::non_NULL(str_gconf_status);

    if (err) {
      vout.crucial(vl, "%s: Input parameters have not been set.\n", test_name.c_str());
      abort();
    }


    // #### Set up a gauge configuration  ####
    Field_G     *U          = new Field_G(Nvol, Ndim);
    GaugeConfig *gconf_read = new GaugeConfig(str_gconf_read);

    if (str_gconf_status == "Continue") {
      gconf_read->read_file((Field *)U, readfile);
    } else if (str_gconf_status == "Cold_start") {
      U->set_unit();
    } else if (str_gconf_status == "Hot_start") {
      RandomNumbers_Mseries rand(1234567);
      U->set_random(&rand);
    } else {
      vout.crucial(vl, "%s: unsupported gconf status \"%s\".\n", test_name.c_str(), str_gconf_status.c_str());
      abort();
    }


    // #### object setup #####
    TopologicalCharge *topological_charge = new TopologicalCharge();
    topological_charge->set_parameters(*params_topo);

    Projection *proj  = Projection::New(str_proj_type);
    Smear      *smear = Smear::New(str_smear_type, proj);
    smear->set_parameters(*params_smear);

    Timer *timer = new Timer(test_name);


    // ####  Execution main part  ####
    timer->start();

    Field_G Uorg(Nvol, Ndim), Usmear(Nvol, Ndim);
    Uorg   = *U;
    Usmear = *U;

    double result = 0.0;
    for (int i_smear = 0; i_smear <= Nsmear; ++i_smear) {
      if (i_smear > 0) smear->smear(Usmear, Uorg);

      if ((i_smear % Nmeas) == 0) {
        result = topological_charge->measure(Usmear);

        vout.general(vl, "i_smear,Q = %d %20.16e\n", i_smear, result);
      }

      Uorg = Usmear;
    }

    timer->report();


    // ####  tydy up  ####
    delete params_test;
    delete params_topo;
    delete params_proj;
    delete params_smear;
    delete params_all;

    delete timer;

    delete U;
    delete gconf_read;

    delete proj;
    delete smear;

    delete topological_charge;

    if (do_check) {
      return Test::verify(expected_result, result);
    } else {
      vout.detailed(vl, "check skipped: expected_result not set.\n\n");
      return EXIT_SKIP;
    }
  }
} // namespace Test_TopologicalCharge
