/*!
        @file    $Id: test_WilsonLoop.cpp #$

        @brief

        @author  Hideo Matsufuru <hideo.matsufuru@kek.jp> (matsufuru)
                 $LastChangedBy: aoym $

        @date    $LastChangedDate: 2013-01-22 13:51:53 #$

        @version $LastChangedRevision: 1144 $
*/

#include "test.h"

#include "gaugeConfig.h"

#include "wilsonLoop.h"

#include "projection_Maximum_SU_N.h"
#include "smear_APE_spatial.h"

#include "randomNumbers_Mseries.h"

//====================================================================
//! Test of Wilson loop measurement.

/*!
                          [27 May 2012 H.Matsufuru]
    (Coding history will be recovered from trac.)
    YAML is implemented.  [14 Nov 2012 Y.Namekawa]
 */

namespace Test_WilsonLoop {
  const std::string test_name = "WilsonLoop";

  //- test-private parameters
  namespace {
    const std::string filename_input  = "test_WilsonLoop.yaml";
    const std::string filename_output = "stdout";

    class Parameters_Test_WilsonLoop : public Parameters {
     public:
      Parameters_Test_WilsonLoop()
      {
        Register_string("gauge_config_status", "NULL");
        Register_string("gauge_config_type_input", "NULL");
        Register_string("config_filename_input", "NULL");

        Register_int("number_of_max_smearing", 0);
        Register_int("number_of_smearing_step", 0);

        Register_string("verbose_level", "NULL");

        Register_double("expected_result", 0.0);
      }
    };
  }

  //- prototype declaration
  int wilsonloop(void);

#ifdef USE_TESTMANAGER_AUTOREGISTER
  namespace {
#if defined(USE_GROUP_SU2)
    // Nc=2 is not available.
#else
    static const bool is_registered = TestManager::RegisterTest(
      test_name,
      wilsonloop
      );
#endif
  }
#endif

  //====================================================================
  int wilsonloop(void)
  {
    // ####  parameter setup  ####
    int Nvol = CommonParameters::Nvol();
    int Ndim = CommonParameters::Ndim();

    Parameters *params_test       = new Parameters_Test_WilsonLoop;
    Parameters *params_wilsonloop = ParametersFactory::New("WilsonLoop");
    Parameters *params_smear      = ParametersFactory::New("Smear.APE_Spatial");
    Parameters *params_all        = new Parameters;

    params_all->Register_Parameters("Test_WilsonLoop", params_test);
    params_all->Register_Parameters("WilsonLoop", params_wilsonloop);
    params_all->Register_Parameters("Smear_APE_spatial", params_smear);

    ParameterManager_YAML params_manager;
    params_manager.read_params(filename_input, params_all);

    const string str_gconf_status = params_test->get_string("gauge_config_status");
    const string str_gconf_read   = params_test->get_string("gauge_config_type_input");
    const string readfile         = params_test->get_string("config_filename_input");
    const int    Nsmear           = params_test->get_int("number_of_max_smearing");
    const int    Nmeas            = params_test->get_int("number_of_smearing_step");
    const string str_vlevel       = params_test->get_string("verbose_level");

    const bool   do_check        = params_test->is_set("expected_result");
    const double expected_result = do_check ? params_test->get_double("expected_result") : 0.0;

    Bridge::VerboseLevel vl = vout.set_verbose_level(str_vlevel);

    //- print input parameters
    vout.general(vl, "  gconf_status = %s\n", str_gconf_status.c_str());
    vout.general(vl, "  gconf_read   = %s\n", str_gconf_read.c_str());
    vout.general(vl, "  readfile     = %s\n", readfile.c_str());
    vout.general(vl, "  Nsmear       = %d\n", Nsmear);
    vout.general(vl, "  Nmeas        = %d\n", Nmeas);
    vout.general(vl, "  vlevel       = %s\n", str_vlevel.c_str());
    vout.general(vl, "\n");

    //- input parameter check
    int err = 0;
    err += ParameterCheck::non_NULL(str_gconf_status);

    if (err) {
      vout.crucial(vl, "%s: Input parameters have not been set.\n", test_name.c_str());
      abort();
    }


    // #### Set up a gauge configuration  ####
    Field_G     *U          = new Field_G(Nvol, Ndim);
    GaugeConfig *gconf_read = new GaugeConfig(str_gconf_read);

    if (str_gconf_status == "Continue") {
      gconf_read->read_file((Field *)U, readfile);
    } else if (str_gconf_status == "Cold_start") {
      U->set_unit();
    } else if (str_gconf_status == "Hot_start") {
      RandomNumbers_Mseries rand(1234567);
      U->set_random(&rand);
    } else {
      vout.crucial(vl, "%s: unsupported gconf status \"%s\".\n", test_name.c_str(), str_gconf_status.c_str());
      abort();
    }


    // #### object setup #####
    WilsonLoop wilsonloop;
    wilsonloop.set_parameters(*params_wilsonloop);

    Projection        *proj  = new Projection_Maximum_SU_N();
    Smear_APE_spatial *smear = new Smear_APE_spatial(proj);
    smear->set_parameters(*params_smear);

    Timer *timer = new Timer(test_name);


    // ####  Execution main part  ####
    timer->start();

    Field_G Uorg(Nvol, Ndim), Usmear(Nvol, Ndim);
    Uorg   = *U;
    Usmear = *U;

    double result = 0.0;
    for (int i_smear = 0; i_smear <= Nsmear; ++i_smear) {
      vout.general(vl, "i_smear = %d\n", i_smear);

      if (i_smear > 0) smear->smear(Usmear, Uorg);

      if ((i_smear % Nmeas) == 0) {
        result = wilsonloop.measure(Usmear);
      }

      Uorg = Usmear;
    }

    timer->report();


    // ####  tydy up  ####
    delete params_test;
    delete params_wilsonloop;
    delete params_smear;
    delete params_all;

    delete timer;

    delete U;
    delete gconf_read;
    delete proj;
    delete smear;


    if (do_check) {
      return Test::verify(expected_result, result);
    } else {
      vout.detailed(vl, "check skipped: expected_result not set.\n\n");
      return EXIT_SKIP;
    }
  }
} // namespace Test_WilsonLoop
